#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020, 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.hy2_scat_l2b_h5 module."""

import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(np.float32)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(-10, 10, DEFAULT_FILE_SHAPE[1]).astype(np.float32)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def __getitem__(self, key):
        """Return copy of dataarray to prevent manipulating attributes in the original."""
        val = self.file_content[key]
        if isinstance(val, xr.core.dataarray.DataArray):
            val = val.copy()
        return val

    def _get_geo_data(self, num_rows, num_cols):
        geo = {
            "wvc_lon":
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        "fill_value": 1.7e+38,
                        "scale_factor": 1.,
                        "add_offset": 0.,
                        "units": "degree",
                        "valid range": [0, 359.99],
                    },
                    dims=("y", "x")),
            "wvc_lat":
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        "fill_value": 1.7e+38,
                        "scale_factor": 1.,
                        "add_offset": 0.,
                        "units": "degree",
                        "valid range": [-90.0, 90.0],
                    },
                    dims=("y", "x")),
        }
        return geo

    def _get_geo_data_nsoas(self, num_rows, num_cols):
        geo = {
            "wvc_lon":
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        "fill_value": 1.7e+38,
                        "scale_factor": 1.,
                        "add_offset": 0.,
                        "units": "degree",
                        "valid_range": [0, 359.99],
                    },
                    dims=("y", "x")),
            "wvc_lat":
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        "fill_value": 1.7e+38,
                        "scale_factor": 1.,
                        "add_offset": 0.,
                        "units": "degree",
                        "valid_range": [-90.0, 90.0],
                    },
                    dims=("y", "x")),
        }
        return geo

    def _get_selection_data(self, num_rows, num_cols):
        selection = {
            "wvc_selection":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 8],
                },
                dims=("y", "x")),
            "wind_speed_selection":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    "fill_value": -32767,
                    "scale_factor": 0.1,
                    "add_offset": 0.,
                    "units": "deg",
                    "valid range": [0, 3599],
                },
                dims=("y", "x")),
            "wind_dir_selection":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    "fill_value": -32767,
                    "scale_factor": 0.01,
                    "add_offset": 0.,
                    "units": "m/s",
                    "valid range": [0, 5000],
                },
                dims=("y", "x")),
            "model_dir":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    "fill_value": -32767,
                    "scale_factor": 0.01,
                    "add_offset": 0.,
                    "units": "m/s",
                    "valid range": [0, 5000],
                },
                dims=("y", "x")),
            "model_speed":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    "fill_value": -32767,
                    "scale_factor": 0.1,
                    "add_offset": 0.,
                    "units": "deg",
                    "valid range": [0, 3599],
                },
                dims=("y", "x")),
            "num_ambigs":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 8],
                },
                dims=("y", "x")),
            "num_in_aft":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 127],
                },
                dims=("y", "x")),
            "num_in_fore":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 127],
                },
                dims=("y", "x")),
            "num_out_aft":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 127],
                },
                dims=("y", "x")),
            "num_out_fore":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    "fill_value": 0,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "count",
                    "valid range": [1, 127],
                },
                dims=("y", "x")),
            "wvc_quality_flag":
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.uint16),
                attrs={
                    "fill_value": 2.14748e+09,
                    "scale_factor": 1.,
                    "add_offset": 0.,
                    "units": "na",
                    "valid range": [1, 2.14748e+09],
                },
                dims=("y", "x")),
        }
        return selection

    def _get_all_ambiguities_data(self, num_rows, num_cols, num_amb):
        all_amb = {
            "max_likelihood_est":
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        "fill_value": -32767,
                        "scale_factor": 1.,
                        "add_offset": 0.,
                        "units": "na",
                        "valid range": [0, 32767],
                    },
                    dims=("y", "x", "selection")),
            "wind_dir":
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        "fill_value": -32767,
                        "scale_factor": 0.1,
                        "add_offset": 0.,
                        "units": "deg",
                        "valid range": [0, 3599],
                    },
                    dims=("y", "x", "selection")),
            "wind_speed":
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        "fill_value": -32767,
                        "scale_factor": 0.01,
                        "add_offset": 0.,
                        "units": "m/s",
                        "valid range": [0, 5000],
                    },
                    dims=("y", "x", "selection")),
        }
        return all_amb

    def _get_wvc_row_time(self, num_rows):
        data = ["20200326T01:11:07.639",
                "20200326T01:11:11.443",
                "20200326T01:11:15.246",
                "20200326T01:11:19.049",
                "20200326T01:11:22.856",
                "20200326T01:11:26.660",
                "20200326T01:11:30.464",
                "20200326T01:11:34.268",
                "20200326T01:11:38.074",
                "20200326T01:11:41.887"]
        wvc_row_time = {
            "wvc_row_time":
                xr.DataArray(data,
                             attrs={
                                 "fill_value": "",
                             },
                             dims=("y",)),
        }
        return wvc_row_time

    def _get_global_attrs(self, num_rows, num_cols):
        return {
            "/attr/Equator_Crossing_Longitude": "246.408397",
            "/attr/Equator_Crossing_Time": "20200326T01:37:15.875",
            "/attr/HDF_Version_Id": "HDF5-1.8.16",
            "/attr/Input_L2A_Filename": "H2B_OPER_SCA_L2A_OR_20200326T010839_20200326T025757_07076_dps_250_20.h5",
            "/attr/Instrument_ShorName": "HSCAT-B",
            "/attr/L2A_Inputdata_Version": "10",
            "/attr/L2B_Actual_WVC_Rows": np.int32(num_rows),
            "/attr/L2B_Algorithm_Descriptor": ("Wind retrieval processing uses the multiple solution scheme (MSS) for "
                                               "wind inversion with the NSCAT-4 GMF,and a circular median filter "
                                               "method (CMF) for ambiguity removal. The ECMWF/NCEP forescate data are "
                                               "used as background winds in the CMF"),
            "/attr/L2B_Data_Version": "10",
            "/attr/L2B_Expected_WVC_Rows": np.int32(num_rows),
            "/attr/L2B_Processing_Type": "OPER",
            "/attr/L2B_Processor_Name": "hy2_sca_l2b_pro",
            "/attr/L2B_Processor_Version": "01.00",
            "/attr/Long_Name": "HY-2B/SCAT Level 2B Ocean Wind Vectors in 25.0 km Swath Grid",
            "/attr/Orbit_Inclination": np.float32(99.3401),
            "/attr/Orbit_Number": "07076",
            "/attr/Output_L2B_Filename": "H2B_OPER_SCA_L2B_OR_20200326T011107_20200326T025540_07076_dps_250_20_owv.h5",
            "/attr/Platform_LongName": "Haiyang 2B Ocean Observing Satellite",
            "/attr/Platform_ShortName": "HY-2B",
            "/attr/Platform_Type": "spacecraft",
            "/attr/Producer_Agency": "Ministry of Natural Resources of the People's Republic of China",
            "/attr/Producer_Institution": "NSOAS",
            "/attr/Production_Date_Time": "20200326T06:23:10",
            "/attr/Range_Beginning_Time": "20200326T01:11:07",
            "/attr/Range_Ending_Time": "20200326T02:55:40",
            "/attr/Rev_Orbit_Period": "14 days",
            "/attr/Short_Name": "HY-2B SCAT-L2B-25km",
            "/attr/Sigma0_Granularity": "whole pulse",
            "/attr/WVC_Size": "25000m*25000m",
        }

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        num_rows = 300
        num_cols = 10
        num_amb = 8

        test_content = {}
        test_content.update(self._get_global_attrs(num_rows, num_cols))
        data = {}
        if "OPER_SCA_L2B" in filename:
            test_content.update({"/attr/L2B_Expected_WVC_Cells": np.int32(num_cols)})
            data = self._get_geo_data_nsoas(num_rows, num_cols)
        else:
            test_content.update({"/attr/L2B_Number_WVC_cells": np.int32(num_cols)})
            data = self._get_geo_data(num_rows, num_cols)

        test_content.update(data)
        data = self._get_selection_data(num_rows, num_cols)
        test_content.update(data)
        data = self._get_all_ambiguities_data(num_rows, num_cols, num_amb)
        test_content.update(data)

        data = self._get_wvc_row_time(num_rows)
        test_content.update(data)

        return test_content


class TestHY2SCATL2BH5Reader(unittest.TestCase):
    """Test HY2 Scatterometer L2B H5 Reader."""

    yaml_file = "hy2_scat_l2b_h5.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.hy2_scat_l2b_h5 import HY2SCATL2BH5FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(HY2SCATL2BH5FileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_load_geo(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers

        res = reader.load(["wvc_lon", "wvc_lat"])
        assert 2 == len(res)

    def test_load_geo_nsoas(self):
        """Test loading data from nsoas file."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "H2B_OPER_SCA_L2B_OR_20210803T100304_20210803T104601_13905_pwp_250_07_owv.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers

        res = reader.load(["wvc_lon", "wvc_lat"])
        assert 2 == len(res)

    def test_load_data_selection(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        res = reader.load(["wind_speed_selection",
                           "wind_dir_selection",
                           "wvc_selection"])
        assert 3 == len(res)

    def test_load_data_all_ambiguities(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        res = reader.load(["wind_speed",
                           "wind_dir",
                           "max_likelihood_est",
                           "model_dir",
                           "model_speed",
                           "num_ambigs",
                           "num_in_aft",
                           "num_in_fore",
                           "num_out_aft",
                           "num_out_fore",
                           "wvc_quality_flag"])
        assert 11 == len(res)

    def test_load_data_row_times(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        res = reader.load(["wvc_row_time"])
        assert 1 == len(res)

    def test_reading_attrs(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        reader.create_filehandlers(files)
        # Make sure we have some files
        res = reader.load(["wvc_lon"])
        assert res["wvc_lon"].attrs["L2B_Number_WVC_cells"] == 10
        with pytest.raises(KeyError):
            assert res["wvc_lon"].attrs["L2B_Expected_WVC_Cells"] == 10

    def test_reading_attrs_nsoas(self):
        """Test loading data."""
        from satpy.readers.core.loading import load_reader
        filenames = [
            "H2B_OPER_SCA_L2B_OR_20210803T100304_20210803T104601_13905_pwp_250_07_owv.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        reader.create_filehandlers(files)
        # Make sure we have some files
        res = reader.load(["wvc_lon"])
        with pytest.raises(KeyError):
            assert res["wvc_lon"].attrs["L2B_Number_WVC_cells"] == 10
        assert res["wvc_lon"].attrs["L2B_Expected_WVC_Cells"] == 10

    def test_properties(self):
        """Test platform_name."""
        import datetime as dt

        from satpy.readers.core.loading import load_reader
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5", ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        reader.create_filehandlers(files)
        # Make sure we have some files
        res = reader.load(["wvc_lon"])
        assert res["wvc_lon"].platform_name == "HY-2B"
        assert res["wvc_lon"].start_time == dt.datetime(2020, 3, 26, 1, 11, 7)
        assert res["wvc_lon"].end_time == dt.datetime(2020, 3, 26, 2, 55, 40)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for the FIDUCEO MVIRI FCDR Reader."""

from __future__ import annotations

import os
from functools import partial
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyproj import CRS
from pyresample.geometry import AreaDefinition

from satpy.readers.mviri_l1b_fiduceo_nc import (
    ALTITUDE,
    EQUATOR_RADIUS,
    POLE_RADIUS,
    FiduceoMviriEasyFcdrFileHandler,
    FiduceoMviriFullFcdrFileHandler,
    Interpolator,
    ignore_dup_dim_warning,
    preprocess_dataset,
)
from satpy.tests.utils import make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - request

fill_val = np.uint32(429496729) # FillValue lower than in dataset to be windows-compatible

attrs_exp: dict = {
    "platform": "MET7",
    "raw_metadata": {"foo": "bar"},
    "sensor": "MVIRI",
    "orbital_parameters": {
        "projection_longitude": 57.0,
        "projection_latitude": 0.0,
        "projection_altitude": 35785860.0,
        "satellite_actual_longitude": 57.1,
        "satellite_actual_latitude": 0.1,
    }

}
attrs_refl_exp = attrs_exp.copy()
attrs_refl_exp.update(
    {"sun_earth_distance_correction_applied": True,
     "sun_earth_distance_correction_factor": 1.}
)
acq_time_vis_exp = [np.datetime64("NaT").astype("datetime64[ns]"),
                    np.datetime64("NaT").astype("datetime64[ns]"),
                    np.datetime64("1970-01-01 02:30").astype("datetime64[ns]"),
                    np.datetime64("1970-01-01 02:30").astype("datetime64[ns]")]
vis_counts_exp = xr.DataArray(
    np.array(
        [[0., 17., 34., 51.],
         [68., 85., 102., 119.],
         [136., 153., np.nan, 187.],
         [204., 221., 238., 255]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_vis_exp),
    },
    attrs=attrs_exp
)

vis_rad_exp = xr.DataArray(
    np.array(
        [[np.nan, 18.56, 38.28, 58.],
         [77.72, 97.44, 117.16, 136.88],
         [156.6, 176.32, np.nan, 215.76],
         [235.48, 255.2, 274.92, 294.64]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_vis_exp),
    },
    attrs=attrs_exp
)
vis_refl_exp = xr.DataArray(
    np.array(
        [[np.nan, 23.440929, np.nan, np.nan],
         [40.658744, 66.602233, 147.970867, np.nan],
         [75.688217, 92.240733, np.nan, np.nan],
         [np.nan, np.nan, np.nan, np.nan]],
        dtype=np.float32
    ),
    # (0, 0) and (2, 2) are NaN because radiance is NaN
    # (0, 2) is NaN because SZA >= 90 degrees
    # Last row/col is NaN due to SZA interpolation
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_vis_exp),
    },
    attrs=attrs_refl_exp
)
u_vis_refl_exp = xr.DataArray(
    np.array(
        [[0.1, 0.2, 0.3, 0.4],
         [0.5, 0.6, 0.7, 0.8],
         [0.9, 1.0, 1.1, 1.2],
         [1.3, 1.4, 1.5, 1.6]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_vis_exp),
    },
    attrs=attrs_exp
)

u_struct_refl_exp = u_vis_refl_exp.copy()

acq_time_ir_wv_exp = [np.datetime64("NaT"),
                      np.datetime64("1970-01-01 02:30").astype("datetime64[ns]")]
wv_counts_exp = xr.DataArray(
    np.array(
        [[0, 85],
         [170, 255]],
        dtype=np.uint8
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
wv_rad_exp = xr.DataArray(
    np.array(
        [[np.nan, 3.75],
         [8, 12.25]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
wv_bt_exp = xr.DataArray(
    np.array(
        [[np.nan, 230.461366],
         [252.507448, 266.863289]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
ir_counts_exp = xr.DataArray(
    np.array(
        [[0, 85],
         [170, 255]],
        dtype=np.uint8
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
ir_rad_exp = xr.DataArray(
    np.array(
        [[np.nan, 80],
         [165, 250]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
ir_bt_exp = xr.DataArray(
    np.array(
        [[np.nan, 178.00013189],
         [204.32955838, 223.28709913]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_ir_wv_exp),
    },
    attrs=attrs_exp
)
quality_pixel_bitmask_exp = xr.DataArray(
    np.array(
        [[0, 0, 0, 0],
         [0, 0, 0, 0],
         [0, 0, 1, 0],
         [0, 0, 0, 0]],
        dtype=np.uint8
    ),
    dims=("y", "x"),
    coords={
        "acq_time": ("y", acq_time_vis_exp),
    },
    attrs=attrs_exp
)
sza_vis_exp = xr.DataArray(
    np.array(
        [[45., 67.5, 90., np.nan],
         [22.5, 45., 67.5, np.nan],
         [0., 22.5, 45., np.nan],
         [np.nan, np.nan, np.nan, np.nan]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    attrs=attrs_exp
)
sza_ir_wv_exp = xr.DataArray(
    np.array(
        [[45, 90],
         [0, 45]],
        dtype=np.float32
    ),
    dims=("y", "x"),
    attrs=attrs_exp
)
projection = CRS(f"+proj=geos +lon_0=57.0 +h={ALTITUDE} +a={EQUATOR_RADIUS} +b={POLE_RADIUS}")
area_vis_exp = AreaDefinition(
    area_id="geos_mviri_4x4",
    proj_id="geos_mviri_4x4",
    description="MVIRI Geostationary Projection",
    projection=projection,
    width=4,
    height=4,
    area_extent=[5621229.74392, 5621229.74392, -5621229.74392, -5621229.74392]
)
area_ir_wv_exp = area_vis_exp.copy(
    area_id="geos_mviri_2x2",
    proj_id="geos_mviri_2x2",
    width=2,
    height=2
)


@pytest.fixture(name="time_fake_dataset")
def fixture_time_fake_dataset():
    """Create time for fake dataset."""
    time = np.arange(4) * 60 * 60
    time[0] = fill_val
    time[1] = fill_val
    time = time.reshape(2, 2)

    return time


@pytest.fixture(name="fake_dataset")
def fixture_fake_dataset(time_fake_dataset):
    """Create fake dataset."""
    count_ir = da.linspace(0, 255, 4, dtype=np.uint8).reshape(2, 2)
    count_wv = da.linspace(0, 255, 4, dtype=np.uint8).reshape(2, 2)
    count_vis = da.linspace(0, 255, 16, dtype=np.uint8).reshape(4, 4)
    sza = da.from_array(
        np.array(
            [[45, 90], [0, 45]],
            dtype=np.float32
        )
    )
    mask = da.from_array(
        np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]], # 1 = "invalid"
            dtype=np.uint8
        )
    )

    cov = da.from_array([[1, 2], [3, 4]])

    with ignore_dup_dim_warning(), xr.set_options(keep_attrs=False):
        ds = xr.Dataset(
            data_vars={
                "count_vis": (("y", "x"), count_vis),
                "count_wv": (("y_ir_wv", "x_ir_wv"), count_wv),
                "count_ir": (("y_ir_wv", "x_ir_wv"), count_ir),
                "toa_bidirectional_reflectance_vis": vis_refl_exp / 100,
                "u_independent_toa_bidirectional_reflectance": u_vis_refl_exp / 100,
                "u_structured_toa_bidirectional_reflectance": u_vis_refl_exp / 100,
                "quality_pixel_bitmask": (("y", "x"), mask),
                "solar_zenith_angle": (("y_tie", "x_tie"), sza),
                "time_ir_wv": (("y_ir_wv", "x_ir_wv"), time_fake_dataset),
                "a_ir": -5.0,
                "b_ir": 1.0,
                "bt_a_ir": 10.0,
                "bt_b_ir": -1000.0,
                "a_wv": -0.5,
                "b_wv": 0.05,
                "bt_a_wv": 10.0,
                "bt_b_wv": -2000.0,
                "years_since_launch": 20.0,
                "a0_vis": 1.0,
                "a1_vis": 0.01,
                "a2_vis": -0.0001,
                "mean_count_space_vis": 1.0,
                "distance_sun_earth": 1.0,
                "solar_irradiance_vis": 650.0,
                "sub_satellite_longitude_start": 57.1,
                "sub_satellite_longitude_end": np.nan,
                "sub_satellite_latitude_start": np.nan,
                "sub_satellite_latitude_end": 0.1,
                "covariance_spectral_response_function_vis": (("srf_size", "srf_size"), cov),
                "channel_correlation_matrix_independent": (("channel", "channel"), cov),
                "channel_correlation_matrix_structured": (("channel", "channel"), cov)
            },
            coords={
                "y": [1, 2, 3, 4],
                "x": [1, 2, 3, 4],
                "y_ir_wv": [1, 2],
                "x_ir_wv": [1, 2],
                "y_tie": [1, 2],
                "x_tie": [1, 2],
            },
            attrs={"foo": "bar"}
        )
    ds["count_ir"].attrs["ancillary_variables"] = "a_ir b_ir"
    ds["count_wv"].attrs["ancillary_variables"] = "a_wv b_wv"
    ds["quality_pixel_bitmask"].encoding["chunksizes"] = (2, 2)
    ds["time_ir_wv"].attrs["_FillValue"] = fill_val
    ds["time_ir_wv"].attrs["add_offset"] = 0

    return ds


@pytest.fixture(name="projection_longitude", params=["57.0"])
def fixture_projection_longitude(request):
    """Get projection longitude as string."""
    return request.param


@pytest.fixture(name="fake_file")
def fixture_fake_file(fake_dataset, tmp_path):
    """Create test file."""
    filename = tmp_path / "test_mviri_fiduceo.nc"
    with ignore_dup_dim_warning():
        fake_dataset.to_netcdf(filename)
    return filename


@pytest.fixture(
    name="file_handler",
    params=[FiduceoMviriEasyFcdrFileHandler,
            FiduceoMviriFullFcdrFileHandler]
)
def fixture_file_handler(fake_file, request, projection_longitude):
    """Create mocked file handler."""
    fh_class = request.param
    return partial(
        fh_class,
        filename=fake_file,
        filename_info={"platform": "MET7",
                       "sensor": "MVIRI",
                       "projection_longitude": projection_longitude},
        filetype_info={"foo": "bar"},
        mask_bad_quality=True,
    )


@pytest.fixture(name="reader")
def fixture_reader():
    """Return MVIRI FIDUCEO FCDR reader."""
    from satpy._config import config_search_paths
    from satpy.readers.core.loading import load_reader

    reader_configs = config_search_paths(
        os.path.join("readers", "mviri_l1b_fiduceo_nc.yaml"))
    reader = load_reader(reader_configs)
    return reader


class TestFiduceoMviriFileHandlers:
    """Unit tests for FIDUCEO MVIRI file handlers."""

    @pytest.mark.parametrize("projection_longitude", ["57.0", "5700"], indirect=True)
    def test_init(self, file_handler, projection_longitude):
        """Test file handler initialization."""
        fh = file_handler()
        assert fh.projection_longitude == 57.0
        assert fh.mask_bad_quality is True

    @pytest.mark.parametrize(
        ("name", "calibration", "resolution", "expected"),
        [
            ("VIS", "counts", 2250, vis_counts_exp),
            ("VIS", "radiance", 2250, vis_rad_exp),
            ("VIS", "reflectance", 2250, vis_refl_exp),
            ("WV", "counts", 4500, wv_counts_exp),
            ("WV", "radiance", 4500, wv_rad_exp),
            ("WV", "brightness_temperature", 4500, wv_bt_exp),
            ("IR", "counts", 4500, ir_counts_exp),
            ("IR", "radiance", 4500, ir_rad_exp),
            ("IR", "brightness_temperature", 4500, ir_bt_exp),
            ("quality_pixel_bitmask", None, 2250, quality_pixel_bitmask_exp),
            ("solar_zenith_angle", None, 2250, sza_vis_exp),
            ("solar_zenith_angle", None, 4500, sza_ir_wv_exp),
            ("u_independent_toa_bidirectional_reflectance", None, 4500, u_vis_refl_exp),
            ("u_structured_toa_bidirectional_reflectance", None, 4500, u_struct_refl_exp)
        ]
    )
    def test_get_dataset(self, file_handler, name, calibration, resolution,
                         expected):
        """Test getting datasets."""
        fh = file_handler()
        id_keys = {"name": name, "resolution": resolution}
        if calibration:
            id_keys["calibration"] = calibration
        dataset_id = make_dataid(**id_keys)
        dataset_info = {"platform": "MET7"}

        is_easy = isinstance(fh, FiduceoMviriEasyFcdrFileHandler)
        is_vis = name == "VIS"
        is_refl = calibration == "reflectance"
        if is_easy and is_vis and not is_refl:
            # VIS counts/radiance not available in easy FCDR
            with pytest.raises(ValueError, match="Cannot calibrate to .*. Easy FCDR provides reflectance only."):
                fh.get_dataset(dataset_id, dataset_info)
        else:
            ds = fh.get_dataset(dataset_id, dataset_info)
            xr.testing.assert_allclose(ds, expected)
            assert ds.dtype == expected.dtype
            assert ds.attrs == expected.attrs

    def test_get_dataset_corrupt(self, file_handler):
        """Test getting datasets with known corruptions."""
        fh = file_handler()
        # Satellite position might be missing
        fh.nc.ds = fh.nc.ds.drop_vars(
            ["sub_satellite_longitude_start"]
        )

        dataset_id = make_dataid(
            name="VIS",
            calibration="reflectance",
            resolution=2250
        )
        ds = fh.get_dataset(dataset_id, {"platform": "MET7"})
        assert "actual_satellite_longitude" not in ds.attrs["orbital_parameters"]
        assert "actual_satellite_latitude" not in ds.attrs["orbital_parameters"]
        xr.testing.assert_allclose(ds, vis_refl_exp)

    @mock.patch(
        "satpy.readers.mviri_l1b_fiduceo_nc.Interpolator.interp_acq_time"
    )
    def test_time_cache(self, interp_acq_time, file_handler):
        """Test caching of acquisition times."""
        fh = file_handler()
        dataset_id = make_dataid(
            name="VIS",
            resolution=2250,
            calibration="reflectance"
        )
        info = {}
        interp_acq_time.return_value = xr.DataArray([1, 2, 3, 4], dims="y")

        # Cache init
        fh.get_dataset(dataset_id, info)
        interp_acq_time.assert_called()

        # Cache hit
        interp_acq_time.reset_mock()
        fh.get_dataset(dataset_id, info)
        interp_acq_time.assert_not_called()

        # Cache miss
        interp_acq_time.return_value = xr.DataArray([1, 2], dims="y")
        another_id = make_dataid(
            name="IR",
            resolution=4500,
            calibration="brightness_temperature"
        )
        interp_acq_time.reset_mock()
        fh.get_dataset(another_id, info)
        interp_acq_time.assert_called()

    @mock.patch(
        "satpy.readers.mviri_l1b_fiduceo_nc.Interpolator.interp_tiepoints"
    )
    def test_angle_cache(self, interp_tiepoints, file_handler):
        """Test caching of angle datasets."""
        fh = file_handler()
        dataset_id = make_dataid(name="solar_zenith_angle",
                                 resolution=2250)
        info = {}

        # Cache init
        fh.get_dataset(dataset_id, info)
        interp_tiepoints.assert_called()

        # Cache hit
        interp_tiepoints.reset_mock()
        fh.get_dataset(dataset_id, info)
        interp_tiepoints.assert_not_called()

        # Cache miss
        another_id = make_dataid(name="solar_zenith_angle",
                                 resolution=4500)
        interp_tiepoints.reset_mock()
        fh.get_dataset(another_id, info)
        interp_tiepoints.assert_called()

    @pytest.mark.parametrize(
        ("name", "resolution", "area_exp"),
        [
            ("VIS", 2250, area_vis_exp),
            ("WV", 4500, area_ir_wv_exp),
            ("IR", 4500, area_ir_wv_exp),
            ("quality_pixel_bitmask", 2250, area_vis_exp),
            ("solar_zenith_angle", 2250, area_vis_exp),
            ("solar_zenith_angle", 4500, area_ir_wv_exp)
        ]
    )
    def test_get_area_definition(self, file_handler, name, resolution,
                                 area_exp):
        """Test getting area definitions."""
        fh = file_handler()
        dataset_id = make_dataid(name=name, resolution=resolution)
        area = fh.get_area_def(dataset_id)

        assert area.crs == area_exp.crs
        np.testing.assert_allclose(area.area_extent, area_exp.area_extent)

    def test_calib_exceptions(self, file_handler):
        """Test calibration exceptions."""
        fh = file_handler()
        with pytest.raises(KeyError):
            fh.get_dataset(
                make_dataid(name="solar_zenith_angle", calibration="counts"),
                {}
            )
        with pytest.raises(KeyError):
            fh.get_dataset(
                make_dataid(
                    name="VIS",
                    resolution=2250,
                    calibration="brightness_temperature"),
                {}
            )
        with pytest.raises(KeyError):
            fh.get_dataset(
                make_dataid(
                    name="IR",
                    resolution=4500,
                    calibration="reflectance"),
                {}
            )
        if isinstance(fh, FiduceoMviriEasyFcdrFileHandler):
            with pytest.raises(KeyError):
                fh.get_dataset(
                    {"name": "VIS", "calibration": "counts"},
                    {}
                )  # not available in easy FCDR

    def test_bad_quality_warning(self, file_handler):
        """Test warning about bad VIS quality."""
        fh = file_handler(mask_bad_quality=False)
        fh.nc.ds["quality_pixel_bitmask"] = 2
        vis = make_dataid(name="VIS", resolution=2250,
                          calibration="reflectance")
        with pytest.warns(UserWarning, match=".*All pixels of the VIS.*"):
            fh.get_dataset(vis, {})

    def test_file_pattern(self, reader):
        """Test file pattern matching."""
        filenames = [
            "FIDUCEO_FCDR_L15_MVIRI_MET7-57.0_201701201000_201701201030_FULL_v2.6_fv3.1.nc",
            "FIDUCEO_FCDR_L15_MVIRI_MET7-57.0_201701201000_201701201030_EASY_v2.6_fv3.1.nc",
            "FIDUCEO_FCDR_L15_MVIRI_MET7-00.0_201701201000_201701201030_EASY_v2.6_fv3.1.nc",
            "MVIRI_FCDR-EASY_L15_MET7-E0000_200607060600_200607060630_0200.nc",
            "MVIRI_FCDR-EASY_L15_MET7-E5700_200607060600_200607060630_0200.nc",
            "MVIRI_FCDR-FULL_L15_MET7-E0000_200607060600_200607060630_0200.nc",
            "abcde",
        ]

        files = reader.select_files_from_pathnames(filenames)
        assert len(files) == 6


class TestDatasetPreprocessor:
    """Test dataset preprocessing."""

    @pytest.fixture(name="dataset")
    def fixture_dataset(self):
        """Get dataset before preprocessing.

        - Encoded timestamps including fill values
        - Duplicate dimension names
        - x/y coordinates not assigned
        """
        time = 60*60
        with ignore_dup_dim_warning():
            ds = xr.Dataset(
                data_vars={
                    "covariance_spectral_response_function_vis": (("srf_size", "srf_size"), [[1, 2], [3, 4]]),
                    "channel_correlation_matrix_independent": (("channel", "channel"), [[1, 2], [3, 4]]),
                    "channel_correlation_matrix_structured": (("channel", "channel"), [[1, 2], [3, 4]]),
                    "time_ir_wv": (("y", "x"), [[time, fill_val], [time, time]],
                                   {"_FillValue": fill_val, "add_offset": 0})
                }
            )
        return ds

    @pytest.fixture(name="dataset_exp")
    def fixture_dataset_exp(self):
        """Get expected dataset after preprocessing.

        - Timestamps should have been converted to datetime64
        - Time dimension should have been renamed
        - Duplicate dimensions should have been removed
        - x/y coordinates should have been assigned
        """
        time_exp = np.datetime64("1970-01-01 01:00").astype("datetime64[ns]")
        return xr.Dataset(
            data_vars={
                "covariance_spectral_response_function_vis": (("srf_size_1", "srf_size_2"), [[1, 2], [3, 4]]),
                "channel_correlation_matrix_independent": (("channel_1", "channel_2"), [[1, 2], [3, 4]]),
                "channel_correlation_matrix_structured": (("channel_1", "channel_2"), [[1, 2], [3, 4]]),
                "time": (("y", "x"), [[time_exp, np.datetime64("NaT")], [time_exp, time_exp]])
            },
            coords={
                "y": [0, 1],
                "x": [0, 1]
            }
        )

    def test_preprocess(self, dataset, dataset_exp):
        """Test dataset preprocessing."""
        with ignore_dup_dim_warning():
            preprocessed = preprocess_dataset(dataset)
        xr.testing.assert_allclose(preprocessed, dataset_exp)


class TestInterpolator:
    """Unit tests for Interpolator class."""
    @pytest.fixture(name="time_ir_wv")
    def fixture_time_ir_wv(self):
        """Returns time_ir_wv."""
        time_ir_wv = xr.DataArray(
            [
              [np.datetime64("1970-01-01 01:00"), np.datetime64("1970-01-01 02:00")],
              [np.datetime64("1970-01-01 03:00"), np.datetime64("1970-01-01 04:00")],
              [np.datetime64("NaT"), np.datetime64("1970-01-01 06:00")],
              [np.datetime64("NaT"), np.datetime64("NaT")],
            ],
            dims=("y", "x"),
            coords={"y": [1, 3, 5, 7]}
        )
        return time_ir_wv.astype("datetime64[ns]")

    @pytest.fixture(name="acq_time_exp")
    def fixture_acq_time_exp(self):
        """Returns acq_time_vis_exp."""
        vis = xr.DataArray(
            [
                np.datetime64("1970-01-01 01:30"),
                np.datetime64("1970-01-01 01:30"),
                np.datetime64("1970-01-01 03:30"),
                np.datetime64("1970-01-01 03:30"),
                np.datetime64("1970-01-01 06:00"),
                np.datetime64("1970-01-01 06:00"),
                np.datetime64("NaT"),
                np.datetime64("NaT")
            ],
            dims="y",
            coords={"y": [1, 2, 3, 4, 5, 6, 7, 8]}
        )

        ir = xr.DataArray(
            [
                np.datetime64("1970-01-01 01:30"),
                np.datetime64("1970-01-01 03:30"),
                np.datetime64("1970-01-01 06:00"),
                np.datetime64("NaT"),
            ],
            dims="y",
            coords={"y": [1, 3, 5, 7]}
        )

        return vis, ir

    def test_interp_acq_time(self, time_ir_wv, acq_time_exp):
        """Tests time interpolation."""
        res_vis = Interpolator.interp_acq_time(time_ir_wv, target_y=acq_time_exp[0].coords["y"])
        res_ir = Interpolator.interp_acq_time(time_ir_wv, target_y=acq_time_exp[1].coords["y"])

        xr.testing.assert_allclose(res_vis, acq_time_exp[0])
        xr.testing.assert_allclose(res_ir, acq_time_exp[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for AMSR L2 reader."""

import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 30)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            "/attr/PlatformShortName": "GCOM-W1",
            "/attr/SensorShortName": "AMSR2",
            "/attr/StartOrbitNumber": "22210",
            "/attr/StopOrbitNumber": "22210",
        }
        k = "Geophysical Data"
        file_content[k] = DEFAULT_FILE_DATA[:, :]
        file_content[k + "/shape"] = (DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])
        file_content[k + "/attr/UNIT"] = "K"
        file_content[k + "/attr/SCALE FACTOR"] = 1

        k = "Latitude of Observation Point"
        file_content[k] = DEFAULT_FILE_DATA[:, :]
        file_content[k + "/shape"] = (DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])
        file_content[k + "/attr/UNIT"] = "deg"
        file_content[k + "/attr/SCALE FACTOR"] = 1
        k = "Longitude of Observation Point"
        file_content[k] = DEFAULT_FILE_DATA[:, :]
        file_content[k + "/shape"] = (DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])
        file_content[k + "/attr/UNIT"] = "deg"
        file_content[k + "/attr/SCALE FACTOR"] = 1

        convert_file_content_to_data_array(file_content, dims=("dim_0", "dim_1"))
        return file_content


class TestAMSR2L2Reader(unittest.TestCase):
    """Test AMSR2 L2 Reader."""

    yaml_file = "amsr2_l2.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.amsr2_l1b import AMSR2L1BFileHandler
        from satpy.readers.amsr2_l2 import AMSR2L2FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(AMSR2L2FileHandler, "__bases__", (FakeHDF5FileHandler2,
                                                                     AMSR2L1BFileHandler))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GW1AM2_202004160129_195B_L2SNSSWLB3300300.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_basic(self):
        """Test loading of basic channels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "GW1AM2_202004160129_195B_L2SNSSWLB3300300.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        ds = r.load(["ssw"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.shape == (DEFAULT_FILE_SHAPE[0], int(DEFAULT_FILE_SHAPE[1]))
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert d.attrs["area"].lons.shape == (DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])
            assert d.attrs["area"].lats.shape == (DEFAULT_FILE_SHAPE[0], DEFAULT_FILE_SHAPE[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Base classes and utilities for all readers configured by YAML files."""

import glob
import itertools
import logging
import os
import warnings
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, deque
from contextlib import suppress
from fnmatch import fnmatch
from weakref import WeakValueDictionary

import numpy as np
import xarray as xr
import yaml
from pyresample.boundary import AreaDefBoundary, Boundary
from pyresample.geometry import AreaDefinition, StackedAreaDefinition, SwathDefinition
from trollsift.parser import globify, parse

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader  # type: ignore

from satpy import DatasetDict
from satpy._compat import cache
from satpy.area import get_area_def
from satpy.aux_download import DataDownloadMixin
from satpy.coords import add_crs_xy_coords
from satpy.dataset import DataID, DataQuery, get_key
from satpy.dataset.dataid import default_co_keys_config, default_id_keys_config, get_keys_from_config
from satpy.utils import recursive_dict_update

logger = logging.getLogger(__name__)


def listify_string(something):
    """Take *something* and make it a list.

    *something* is either a list of strings or a string, in which case the
    function returns a list containing the string.
    If *something* is None, an empty list is returned.
    """
    if isinstance(something, str):
        return [something]
    if something is not None:
        return list(something)
    return list()


def _get_filebase(path, pattern):
    """Get the end of *path* of same length as *pattern*."""
    # convert any `/` on Windows to `\\`
    path = os.path.normpath(path)
    # A pattern can include directories
    tail_len = len(pattern.split(os.path.sep))
    return os.path.join(*str(path).split(os.path.sep)[-tail_len:])


def _match_filenames(filenames, pattern):
    """Get the filenames matching *pattern*."""
    matching = set()
    glob_pat = globify(pattern)
    for filename in filenames:
        if fnmatch(_get_filebase(filename, pattern), glob_pat):
            matching.add(filename)

    return matching


def _verify_reader_info_assign_config_files(config, config_files):
    try:
        reader_info = config["reader"]
    except KeyError:
        raise KeyError(
            "Malformed config file {}: missing reader 'reader'".format(
                config_files))
    else:
        reader_info["config_files"] = config_files


def load_yaml_configs(*config_files, loader=Loader):
    """Merge a series of YAML reader configuration files.

    Args:
        *config_files (str): One or more pathnames
            to YAML-based reader configuration files that will be merged
            to create a single configuration.
        loader: Yaml loader object to load the YAML with. Defaults to
            `CLoader` if libyaml is available, `Loader` otherwise.

    Returns: dict
        Dictionary representing the entire YAML configuration with the
        addition of `config['reader']['config_files']` (the list of
        YAML pathnames that were merged).

    """
    config = {}
    logger.debug("Reading %s", str(config_files))
    for config_file in config_files:
        with open(config_file, "r", encoding="utf-8") as fd:
            recursive_dict_update(config, yaml.load(fd, Loader=loader))
    _verify_reader_info_assign_config_files(config, config_files)
    return config


class AbstractYAMLReader(metaclass=ABCMeta):
    """Base class for all readers that use YAML configuration files.

    This class should only be used in rare cases. Its child class
    `FileYAMLReader` should be used in most cases.

    """

    def __init__(self, config_dict):
        """Load information from YAML configuration file about how to read data files."""
        if isinstance(config_dict, str):
            raise ValueError("Passing config files to create a Reader is "
                             "deprecated. Use ReaderClass.from_config_files "
                             "instead.")
        self.config = config_dict
        self.info = self.config["reader"]
        self.name = self.info["name"]
        self.file_patterns = []
        for file_type, filetype_info in self.config["file_types"].items():
            filetype_info.setdefault("file_type", file_type)
            # correct separator if needed
            file_patterns = [os.path.join(*pattern.split("/"))
                             for pattern in filetype_info["file_patterns"]]
            filetype_info["file_patterns"] = file_patterns
            self.file_patterns.extend(file_patterns)

        if "sensors" in self.info and not isinstance(self.info["sensors"], (list, tuple)):
            self.info["sensors"] = [self.info["sensors"]]
        self.datasets = self.config.get("datasets", {})
        self._id_keys = self.info.get("data_identification_keys", default_id_keys_config)
        self._co_keys = self.info.get("coord_identification_keys", default_co_keys_config)
        self.info["filenames"] = []
        self.all_ids = {}
        self.load_ds_ids_from_config()

    @classmethod
    def from_config_files(cls, *config_files, **reader_kwargs):
        """Create a reader instance from one or more YAML configuration files."""
        config_dict = load_yaml_configs(*config_files)
        return config_dict["reader"]["reader"](config_dict, **reader_kwargs)

    @property
    def sensor_names(self):
        """Names of sensors whose data is being loaded by this reader."""
        return self.info["sensors"] or []

    @property
    def all_dataset_ids(self):
        """Get DataIDs of all datasets known to this reader."""
        return self.all_ids.keys()

    @property
    def all_dataset_names(self):
        """Get names of all datasets known to this reader."""
        # remove the duplicates from various calibration and resolutions
        return set(ds_id["name"] for ds_id in self.all_dataset_ids)

    @property
    def available_dataset_ids(self):
        """Get DataIDs that are loadable by this reader."""
        logger.warning(
            "Available datasets are unknown, returning all datasets...")
        return self.all_dataset_ids

    @property
    def available_dataset_names(self):
        """Get names of datasets that are loadable by this reader."""
        return (ds_id["name"] for ds_id in self.available_dataset_ids)

    @property
    @abstractmethod
    def start_time(self):
        """Start time of the reader."""

    @property
    @abstractmethod
    def end_time(self):
        """End time of the reader."""

    @abstractmethod
    def filter_selected_filenames(self, filenames):
        """Filter provided filenames by parameters in reader configuration.

        Returns: iterable of usable files

        """

    @abstractmethod
    def load(self, dataset_keys):
        """Load *dataset_keys*."""

    def supports_sensor(self, sensor):
        """Check if *sensor* is supported.

        Returns True is *sensor* is None.
        """
        if sensor and not (set(self.info.get("sensors")) &
                           set(listify_string(sensor))):
            return False
        return True

    def select_files_from_directory(
            self, directory=None, fs=None):
        """Find files for this reader in *directory*.

        If directory is None or '', look in the current directory.

        Searches the local file system by default.  Can search on a remote
        filesystem by passing an instance of a suitable implementation of
        ``fsspec.spec.AbstractFileSystem``.

        Args:
            directory (Optional[str]): Path to search.
            fs (Optional[fsspec.spec.AbstractFileSystem]): fsspec FileSystem implementation to use.
                                       Defaults to None, using local file
                                       system.

        Returns:
            list of strings describing matching files
        """
        filenames = set()
        if directory is None:
            directory = ""
        # all the glob patterns that we are going to look at
        all_globs = {os.path.join(directory, globify(pattern))
                     for pattern in self.file_patterns}
        # custom filesystem or not
        if fs is None:
            matcher = glob.iglob
        else:
            matcher = fs.glob
        # get all files matching these patterns
        for glob_pat in all_globs:
            filenames.update(matcher(glob_pat))
        return filenames

    def select_files_from_pathnames(self, filenames):
        """Select the files from *filenames* this reader can handle."""
        selected_filenames = []
        filenames = set(filenames)  # make a copy of the inputs

        for pattern in self.file_patterns:
            matching = _match_filenames(filenames, pattern)
            filenames -= matching
            for fname in matching:
                if fname not in selected_filenames:
                    selected_filenames.append(fname)
        if len(selected_filenames) == 0:
            logger.warning("No filenames found for reader: %s", self.name)
        return selected_filenames

    def get_dataset_key(self, key, **kwargs):
        """Get the fully qualified `DataID` matching `key`.

        See `satpy.readers.get_key` for more information about kwargs.

        """
        return get_key(key, self.all_ids.keys(), **kwargs)

    def load_ds_ids_from_config(self):
        """Get the dataset ids from the config."""
        ids = []
        for dataset in self.datasets.values():
            # xarray doesn't like concatenating attributes that are lists
            # https://github.com/pydata/xarray/issues/2060
            if "coordinates" in dataset and \
                    isinstance(dataset["coordinates"], list):
                dataset["coordinates"] = tuple(dataset["coordinates"])
            id_keys = get_keys_from_config(self._id_keys, dataset)

            # Build each permutation/product of the dataset
            id_kwargs = self._build_id_permutations(dataset, id_keys)

            for id_params in itertools.product(*id_kwargs):
                dsid = DataID(id_keys, **dict(zip(id_keys, id_params)))
                ids.append(dsid)

                # create dataset infos specifically for this permutation
                ds_info = dataset.copy()
                for key in dsid.keys():
                    if isinstance(ds_info.get(key), dict):
                        with suppress(KeyError):
                            # KeyError is suppressed in case the key does not represent interesting metadata,
                            # eg a custom type
                            ds_info.update(ds_info[key][dsid.get(key)])
                    # this is important for wavelength which was converted
                    # to a tuple
                    ds_info[key] = dsid.get(key)
                self.all_ids[dsid] = ds_info
        return ids

    def _build_id_permutations(self, dataset, id_keys):
        """Build each permutation/product of the dataset."""
        id_kwargs = []
        for key, idval in id_keys.items():
            val = dataset.get(key, idval.get("default") if idval is not None else None)
            val_type = None
            if idval is not None:
                val_type = idval.get("type")
            if val_type is not None and issubclass(val_type, tuple):
                # special case: wavelength can be [min, nominal, max]
                # but is still considered 1 option
                id_kwargs.append((val,))
            elif isinstance(val, (list, tuple, set)):
                # this key has multiple choices
                # (ex. 250 meter, 500 meter, 1000 meter resolutions)
                id_kwargs.append(val)
            elif isinstance(val, dict):
                id_kwargs.append(val.keys())
            else:
                # this key only has one choice so make it a one
                # item iterable
                id_kwargs.append((val,))
        return id_kwargs


class GenericYAMLReader(AbstractYAMLReader):
    """A Generic YAML-based reader."""

    def __init__(self, config_dict, filter_parameters=None, filter_filenames=True):
        """Set up the yaml reader."""
        super().__init__(config_dict)
        self.filter_parameters = filter_parameters or {}
        self.filter_filenames = self.info.get("filter_filenames", filter_filenames)

    def filter_selected_filenames(self, filenames):
        """Filter provided files based on metadata in the filename."""
        if not isinstance(filenames, set):
            # we perform set operations later on to improve performance
            filenames = set(filenames)
        for _, filetype_info in self.sorted_filetype_items():
            filename_iter = self.filename_items_for_filetype(filenames,
                                                             filetype_info)
            if self.filter_filenames:
                filename_iter = self.filter_filenames_by_info(filename_iter)

            for fn, _ in filename_iter:
                yield fn

    def sorted_filetype_items(self):
        """Sort the instance's filetypes in using order."""
        processed_types = []
        file_type_items = deque(self.config["file_types"].items())
        while len(file_type_items):
            filetype, filetype_info = file_type_items.popleft()

            requirements = filetype_info.get("requires")
            if requirements is not None:
                # requirements have not been processed yet -> wait
                missing = [req for req in requirements
                           if req not in processed_types]
                if missing:
                    file_type_items.append((filetype, filetype_info))
                    continue

            processed_types.append(filetype)
            yield filetype, filetype_info

    @staticmethod
    def filename_items_for_filetype(filenames, filetype_info):
        """Iterate over the filenames matching *filetype_info*."""
        if not isinstance(filenames, set):
            # we perform set operations later on to improve performance
            filenames = set(filenames)
        for pattern in filetype_info["file_patterns"]:
            matched_files = set()
            matches = _match_filenames(filenames, pattern)
            for filename in matches:
                try:
                    filename_info = parse(
                        pattern, _get_filebase(filename, pattern))
                except ValueError:
                    logger.debug("Can't parse %s with %s.", filename, pattern)
                    continue
                matched_files.add(filename)
                yield filename, filename_info
            filenames -= matched_files

    def filter_filenames_by_info(self, filename_items):
        """Filter out file using metadata from the filenames.

        Currently only uses start and end time. If only start time is available
        from the filename, keep all the filename that have a start time before
        the requested end time.
        """
        for filename, filename_info in filename_items:
            fend = filename_info.get("end_time")
            fstart = filename_info.setdefault("start_time", fend)
            if fend and fend < fstart:
                # correct for filenames with 1 date and 2 times
                fend = fend.replace(year=fstart.year,
                                    month=fstart.month,
                                    day=fstart.day)
                filename_info["end_time"] = fend
            if self.metadata_matches(filename_info):
                yield filename, filename_info

    def metadata_matches(self, sample_dict, file_handler=None):
        """Check that file metadata matches filter_parameters of this reader."""
        # special handling of start/end times
        if not self.time_matches(
                sample_dict.get("start_time"), sample_dict.get("end_time")):
            return False
        for key, val in self.filter_parameters.items():
            if key != "area" and key not in sample_dict:
                continue

            if key in ["start_time", "end_time"]:
                continue
            elif key == "area" and file_handler:
                if not self.check_file_covers_area(file_handler, val):
                    logger.info("Filtering out %s based on area",
                                file_handler.filename)
                    break
            elif key in sample_dict and val != sample_dict[key]:
                # don't use this file
                break
        else:
            # all the metadata keys are equal
            return True
        return False

    def time_matches(self, fstart, fend):
        """Check that a file's start and end time mtach filter_parameters of this reader."""
        start_time = self.filter_parameters.get("start_time")
        end_time = self.filter_parameters.get("end_time")
        fend = fend or fstart
        if start_time and fend and fend < start_time:
            return False
        if end_time and fstart and fstart > end_time:
            return False
        return True


class FileYAMLReader(GenericYAMLReader, DataDownloadMixin):
    """Primary reader base class that is configured by a YAML file.

    This class uses the idea of per-file "file handler" objects to read file
    contents and determine what is available in the file. This differs from
    the base :class:`AbstractYAMLReader` which does not depend on individual
    file handler objects. In almost all cases this class should be used over
    its base class and can be used as a reader by itself and requires no
    subclassing.

    """

    # WeakValueDictionary objects must be created at the class level or else
    # dask will not be able to serialize them on a distributed environment
    _coords_cache: WeakValueDictionary = WeakValueDictionary()

    def __init__(self,
                 config_dict,
                 filter_parameters=None,
                 filter_filenames=True,
                 **kwargs):
        """Set up initial internal storage for loading file data."""
        super().__init__(config_dict, filter_parameters, filter_filenames)

        self.file_handlers = {}
        self.available_ids = {}
        self.register_data_files()

    @property
    def sensor_names(self):
        """Names of sensors whose data is being loaded by this reader."""
        if not self.file_handlers:
            return self.info["sensors"]

        file_handlers = (handlers[0] for handlers in
                         self.file_handlers.values())
        sensor_names = set()
        for fh in file_handlers:
            try:
                sensor_names.update(fh.sensor_names)
            except NotImplementedError:
                continue
        if not sensor_names:
            return self.info["sensors"]
        return sorted(sensor_names)

    @property
    def available_dataset_ids(self):
        """Get DataIDs that are loadable by this reader."""
        return self.available_ids.keys()

    @property
    def start_time(self):
        """Start time of the earlier file used by this reader."""
        if not self.file_handlers:
            raise RuntimeError("Start time unknown until files are selected")
        return min(x[0].start_time for x in self.file_handlers.values())

    @property
    def end_time(self):
        """End time of the latest file used by this reader."""
        if not self.file_handlers:
            raise RuntimeError("End time unknown until files are selected")
        return max(x[-1].end_time for x in self.file_handlers.values())

    @staticmethod
    def check_file_covers_area(file_handler, check_area):
        """Check if the file covers the current area.

        If the file doesn't provide any bounding box information or 'area'
        was not provided in `filter_parameters`, the check returns True.
        """
        try:
            gbb = Boundary(*file_handler.get_bounding_box())
        except NotImplementedError as err:
            logger.debug("Bounding box computation not implemented: %s",
                         str(err))
        else:
            abb = AreaDefBoundary(get_area_def(check_area), frequency=1000)

            intersection = gbb.contour_poly.intersection(abb.contour_poly)
            if not intersection:
                return False
        return True

    def find_required_filehandlers(self, requirements, filename_info):
        """Find the necessary file handlers for the given requirements.

        We assume here requirements are available.

        Raises:
            (Exception) `KeyError`: if no handler for the given requirements is available.
            (Exception) `RuntimeError`: if there is a handler for the given requirements,
                                        but it doesn't match the filename info.

        """
        req_fh = []
        filename_info = set(filename_info.items())
        if requirements:
            for requirement in requirements:
                for fhd in self.file_handlers[requirement]:
                    if set(fhd.filename_info.items()).issubset(filename_info):
                        req_fh.append(fhd)
                        break
                else:
                    raise RuntimeError("No matching requirement file of type "
                                       "{}".format(requirement))
                    # break everything and continue to next
                    # filetype!
        return req_fh

    def _new_filehandler_instances(self, filetype_info, filename_items, fh_kwargs=None):
        """Generate new filehandler instances."""
        requirements = filetype_info.get("requires")
        filetype_cls = filetype_info["file_reader"]

        if fh_kwargs is None:
            fh_kwargs = {}

        for filename, filename_info in filename_items:
            try:
                req_fh = self.find_required_filehandlers(requirements,
                                                         filename_info)
            except KeyError as req:
                msg = "No handler for reading requirement {} for {}".format(
                    req, filename)
                warnings.warn(msg, stacklevel=4)
                continue
            except RuntimeError as err:
                warnings.warn(str(err) + " for {}".format(filename), stacklevel=4)
                continue

            yield filetype_cls(filename, filename_info, filetype_info, *req_fh, **fh_kwargs)

    def filter_fh_by_metadata(self, filehandlers):
        """Filter out filehandlers using provide filter parameters."""
        for filehandler in filehandlers:
            filehandler.metadata["start_time"] = filehandler.start_time
            filehandler.metadata["end_time"] = filehandler.end_time
            if self.metadata_matches(filehandler.metadata, filehandler):
                yield filehandler

    def _new_filehandlers_for_filetype(self, filetype_info, filenames, fh_kwargs=None):
        """Create filehandlers for a given filetype."""
        filename_iter = self.filename_items_for_filetype(filenames,
                                                         filetype_info)
        if self.filter_filenames:
            # preliminary filter of filenames based on start/end time
            # to reduce the number of files to open
            filename_iter = self.filter_filenames_by_info(filename_iter)
        filehandler_iter = self._new_filehandler_instances(filetype_info,
                                                           filename_iter,
                                                           fh_kwargs=fh_kwargs)
        filtered_iter = self.filter_fh_by_metadata(filehandler_iter)
        return list(filtered_iter)


    def create_storage_items(self, files, **kwargs):
        """Create the storage items."""
        return self.create_filehandlers(files, **kwargs)

    def create_filehandlers(self, filenames, fh_kwargs=None):
        """Organize the filenames into file types and create file handlers."""
        filenames = list(OrderedDict.fromkeys(filenames))
        logger.debug("Assigning to %s: %s", self.info["name"], filenames)

        self.info.setdefault("filenames", []).extend(filenames)
        filename_set = set(filenames)
        created_fhs = {}
        # load files that we know about by creating the file handlers
        for filetype, filetype_info in self.sorted_filetype_items():
            filehandlers = self._new_filehandlers_for_filetype(filetype_info,
                                                               filename_set,
                                                               fh_kwargs=fh_kwargs)

            if filehandlers:
                created_fhs[filetype] = filehandlers
                self.file_handlers[filetype] = sorted(
                    self.file_handlers.get(filetype, []) + filehandlers,
                    key=lambda fhd: (fhd.start_time, fhd.filename))

        # Update dataset IDs with IDs determined dynamically from the file
        # and/or update any missing metadata that only the file knows.
        # Check if the dataset ID is loadable from that file.
        self.update_ds_ids_from_file_handlers()
        return created_fhs

    def _file_handlers_available_datasets(self):
        """Generate a series of available dataset information.

        This is done by chaining file handler's
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        together. See that method's documentation for more information.

        Returns:
            Generator of (bool, dict) where the boolean tells whether the
            current dataset is available from any of the file handlers. The
            boolean can also be None in the case where no loaded file handler
            is configured to load the dataset. The
            dictionary is the metadata provided either by the YAML
            configuration files or by the file handler itself if it is a new
            dataset. The file handler may have also supplemented or modified
            the information.

        """
        # flatten all file handlers in to one list
        flat_fhs = (fh for fhs in self.file_handlers.values() for fh in fhs)
        id_values = list(self.all_ids.values())
        configured_datasets = ((None, ds_info) for ds_info in id_values)
        for fh in flat_fhs:
            # chain the 'available_datasets' methods together by calling the
            # current file handler's method with the previous ones result
            configured_datasets = fh.available_datasets(configured_datasets=configured_datasets)
        return configured_datasets

    def update_ds_ids_from_file_handlers(self):
        """Add or modify available dataset information.

        Each file handler is consulted on whether or not it can load the
        dataset with the provided information dictionary.
        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for more information.

        """
        avail_datasets = self._file_handlers_available_datasets()
        new_ids = {}
        for is_avail, ds_info in avail_datasets:
            # especially from the yaml config
            coordinates = ds_info.get("coordinates")
            if isinstance(coordinates, list):
                # xarray doesn't like concatenating attributes that are
                # lists: https://github.com/pydata/xarray/issues/2060
                ds_info["coordinates"] = tuple(ds_info["coordinates"])

            ds_info.setdefault("modifiers", tuple())  # default to no mods

            # Create DataID for this dataset
            ds_id = DataID(self._id_keys, **ds_info)
            # all datasets
            new_ids[ds_id] = ds_info
            # available datasets
            # False == we have the file type but it doesn't have this dataset
            # None == we don't have the file type object to ask
            if is_avail:
                self.available_ids[ds_id] = ds_info
        self.all_ids = new_ids

    @staticmethod
    def _load_dataset(dsid, ds_info, file_handlers, dim="y", **kwargs):
        """Load only a piece of the dataset."""
        slice_list = []
        failure = True
        for fh in file_handlers:
            try:
                projectable = fh.get_dataset(dsid, ds_info)
                if projectable is not None:
                    slice_list.append(projectable)
                    failure = False
            except KeyError:
                logger.warning("Failed to load {} from {}".format(dsid, fh),
                               exc_info=True)

        if failure:
            raise KeyError(
                "Could not load {} from any provided files".format(dsid))

        if dim not in slice_list[0].dims:
            return slice_list[0]
        res = xr.concat(slice_list, dim=dim)

        combined_info = file_handlers[0].combine_info(
            [p.attrs for p in slice_list])

        res.attrs = combined_info
        return res

    def _load_dataset_data(self, file_handlers, dsid, **kwargs):
        ds_info = self.all_ids[dsid]
        proj = self._load_dataset(dsid, ds_info, file_handlers, **kwargs)
        # FIXME: areas could be concatenated here
        # Update the metadata
        proj.attrs["start_time"] = file_handlers[0].start_time
        proj.attrs["end_time"] = file_handlers[-1].end_time
        proj.attrs["reader"] = self.name
        return proj

    def _preferred_filetype(self, filetypes):
        """Get the preferred filetype out of the *filetypes* list.

        At the moment, it just returns the first filetype that has been loaded.
        """
        if not isinstance(filetypes, list):
            filetypes = [filetypes]

        # look through the file types and use the first one that we have loaded
        for filetype in filetypes:
            if filetype in self.file_handlers:
                return filetype
        return None

    def _load_area_def(self, dsid, file_handlers, **kwargs):
        """Load the area definition of *dsid*."""
        return _load_area_def(dsid, file_handlers)

    def _get_file_handlers(self, dsid):
        """Get the file handler to load this dataset."""
        ds_info = self.all_ids[dsid]

        filetype = self._preferred_filetype(ds_info["file_type"])
        if filetype is None:
            logger.warning("Required file type '%s' not found or loaded for "
                           "'%s'", ds_info["file_type"], dsid["name"])
        else:
            return self.file_handlers[filetype]

    def _load_dataset_area(self, dsid, file_handlers, coords, **kwargs):
        """Get the area for *dsid*."""
        try:
            return self._load_area_def(dsid, file_handlers, **kwargs)
        except NotImplementedError:
            if any(x is None for x in coords):
                logger.warning(
                    "Failed to load coordinates for '{}'".format(dsid))
                return None

            area = self._make_area_from_coords(coords)
            if area is None:
                logger.debug("No coordinates found for %s", str(dsid))
            return area

    def _make_area_from_coords(self, coords):
        """Create an appropriate area with the given *coords*."""
        if len(coords) == 2:
            lons, lats = self._get_lons_lats_from_coords(coords)
            sdef = self._make_swath_definition_from_lons_lats(lons, lats)
            return sdef
        if len(coords) != 0:
            raise NameError("Don't know what to do with coordinates " + str(
                coords))

    def _get_lons_lats_from_coords(self, coords):
        """Get lons and lats from the coords list."""
        lons, lats = None, None
        for coord in coords:
            if coord.attrs.get("standard_name") == "longitude":
                lons = coord
            elif coord.attrs.get("standard_name") == "latitude":
                lats = coord
        if lons is None or lats is None:
            raise ValueError("Missing longitude or latitude coordinate: " + str(coords))
        return lons, lats

    def _make_swath_definition_from_lons_lats(self, lons, lats):
        """Make a swath definition instance from lons and lats."""
        key = None
        try:
            key = (lons.data.name, lats.data.name)
            sdef = FileYAMLReader._coords_cache.get(key)
        except AttributeError:
            sdef = None
        if sdef is None:
            sdef = SwathDefinition(lons, lats)
            sensor_str = "_".join(self.info["sensors"])
            shape_str = "_".join(map(str, lons.shape))
            sdef.name = "{}_{}_{}_{}".format(sensor_str, shape_str,
                                             lons.attrs.get("name", lons.name),
                                             lats.attrs.get("name", lats.name))
            if key is not None:
                FileYAMLReader._coords_cache[key] = sdef
        return sdef

    def _load_dataset_with_area(self, dsid, coords, **kwargs):
        """Load *dsid* and its area if available."""
        file_handlers = self._get_file_handlers(dsid)
        if not file_handlers:
            return

        try:
            ds = self._load_dataset_data(file_handlers, dsid, **kwargs)
        except (KeyError, ValueError) as err:
            logger.exception("Could not load dataset '%s': %s", dsid, str(err))
            return None

        coords = self._assign_coords_from_dataarray(coords, ds)

        area = self._load_dataset_area(dsid, file_handlers, coords, **kwargs)

        if area is not None:
            ds.attrs["area"] = area
            ds = add_crs_xy_coords(ds, area)
        return ds

    @staticmethod
    def _assign_coords_from_dataarray(coords, ds):
        """Assign coords from the *ds* dataarray if needed."""
        if not coords:
            coords = []
            for coord in ds.coords.values():
                if coord.attrs.get("standard_name") in ["longitude", "latitude"]:
                    coords.append(coord)
        return coords

    def _load_ancillary_variables(self, datasets, **kwargs):
        """Load the ancillary variables of `datasets`."""
        all_av_ids = self._gather_ancillary_variables_ids(datasets)
        loadable_av_ids = [av_id for av_id in all_av_ids if av_id not in datasets]
        if not all_av_ids:
            return
        if loadable_av_ids:
            self.load(loadable_av_ids, previous_datasets=datasets, **kwargs)

        for dataset in datasets.values():
            new_vars = []
            for av_id in dataset.attrs.get("ancillary_variables", []):
                if isinstance(av_id, DataID):
                    new_vars.append(datasets[av_id])
                else:
                    new_vars.append(av_id)
            dataset.attrs["ancillary_variables"] = new_vars

    def _gather_ancillary_variables_ids(self, datasets):
        """Gather ancillary variables' ids.

        This adds/modifies the dataset's `ancillary_variables` attr.
        """
        all_av_ids = set()
        for dataset in datasets.values():
            ancillary_variables = dataset.attrs.get("ancillary_variables", [])
            if not isinstance(ancillary_variables, (list, tuple, set)):
                ancillary_variables = ancillary_variables.split(" ")
            av_ids = []
            for key in ancillary_variables:
                try:
                    av_ids.append(self.get_dataset_key(key))
                except KeyError:
                    logger.warning("Can't load ancillary dataset %s", str(key))

            all_av_ids |= set(av_ids)
            dataset.attrs["ancillary_variables"] = av_ids
        return all_av_ids

    def get_dataset_key(self, key, available_only=False, **kwargs):
        """Get the fully qualified `DataID` matching `key`.

        This will first search through available DataIDs, datasets that
        should be possible to load, and fallback to "known" datasets, those
        that are configured but aren't loadable from the provided files.
        Providing ``available_only=True`` will stop this fallback behavior
        and raise a ``KeyError`` exception if no available dataset is found.

        Args:
            key (str, float, DataID, DataQuery): Key to search for in this reader.
            available_only (bool): Search only loadable datasets for the
                provided key. Loadable datasets are always searched first,
                but if ``available_only=False`` (default) then all known
                datasets will be searched.
            kwargs: See :func:`satpy.dataset.data_dict.get_key` for more information about
                kwargs.

        Returns:
            Best matching DataID to the provided ``key``.

        Raises:
            (Exception) KeyError: if no key match is found.

        """
        try:
            return get_key(key, self.available_dataset_ids, **kwargs)
        except KeyError:
            if available_only:
                raise
            return get_key(key, self.all_dataset_ids, **kwargs)

    def load(self, dataset_keys, previous_datasets=None, **kwargs):
        """Load `dataset_keys`.

        If `previous_datasets` is provided, do not reload those.
        """
        all_datasets = previous_datasets or DatasetDict()
        datasets = DatasetDict()

        # Include coordinates in the list of datasets to load
        dsids = [self.get_dataset_key(ds_key) for ds_key in dataset_keys]
        coordinates = self._get_coordinates_for_dataset_keys(dsids)
        all_dsids = list(set().union(*coordinates.values())) + dsids
        for dsid in all_dsids:
            if dsid in all_datasets:
                continue
            coords = [all_datasets.get(cid, None)
                      for cid in coordinates.get(dsid, [])]
            ds = self._load_dataset_with_area(dsid, coords, **kwargs)
            if ds is not None:
                all_datasets[dsid] = ds
                if dsid in dsids:
                    datasets[dsid] = ds
        self._load_ancillary_variables(all_datasets, **kwargs)

        return datasets

    def _get_coordinates_for_dataset_keys(self, dsids):
        """Get all coordinates."""
        coordinates = {}
        for dsid in dsids:
            cids = self._get_coordinates_for_dataset_key(dsid)
            coordinates.setdefault(dsid, []).extend(cids)
        return coordinates

    def _get_coordinates_for_dataset_key(self, dsid):
        """Get the coordinate dataset keys for *dsid*."""
        ds_info = self.all_ids[dsid]
        cids = []
        for cinfo in ds_info.get("coordinates", []):
            if not isinstance(cinfo, dict):
                cinfo = {"name": cinfo}

            for key in self._co_keys:
                if key == "name":
                    continue
                if key in ds_info:
                    if ds_info[key] is not None:
                        cinfo[key] = ds_info[key]
            cid = DataQuery.from_dict(cinfo)

            cids.append(self.get_dataset_key(cid))

        return cids


def _load_area_def(dsid, file_handlers):
    """Load the area definition of *dsid*."""
    area_defs = [fh.get_area_def(dsid) for fh in file_handlers]
    area_defs = [area_def for area_def in area_defs
                 if area_def is not None]

    final_area = StackedAreaDefinition(*area_defs)
    return final_area.squeeze()


def _set_orientation(dataset, upper_right_corner):
    """Set the orientation of geostationary datasets.

    Allows to flip geostationary imagery when loading the datasets.
    Example call: scn.load(['VIS008'], upper_right_corner='NE')

    Args:
        dataset: Dataset to be flipped.
        upper_right_corner (str): Direction of the upper right corner of the image after flipping.
                                Possible options are 'NW', 'NE', 'SW', 'SE', or 'native'.
                                The common upright image orientation corresponds to 'NE'.
                                Defaults to 'native' (no flipping is applied).

    """
    # do some checks and early returns
    if upper_right_corner == "native":
        logger.debug("Requested orientation for Dataset {} is 'native' (default). "
                     "No flipping is applied.".format(dataset.attrs.get("name")))
        return dataset

    if upper_right_corner not in ["NW", "NE", "SE", "SW", "native"]:
        raise ValueError("Target orientation for Dataset {} not recognized. "
                         "Kwarg upper_right_corner should be "
                         "'NW', 'NE', 'SW', 'SE' or 'native'.".format(dataset.attrs.get("name", "unknown_name")))

    if "area" not in dataset.attrs:
        logger.info("Dataset {} is missing the area attribute "
                    "and will not be flipped.".format(dataset.attrs.get("name", "unknown_name")))
        return dataset

    if isinstance(dataset.attrs["area"], SwathDefinition):
        logger.info("Dataset {} is in a SwathDefinition "
                    "and will not be flipped.".format(dataset.attrs.get("name", "unknown_name")))
        return dataset

    projection_type = _get_projection_type(dataset.attrs["area"])
    accepted_geos_proj_types = ["Geostationary Satellite (Sweep Y)", "Geostationary Satellite (Sweep X)"]
    if projection_type not in accepted_geos_proj_types:
        logger.info("Dataset {} is not in one of the known geostationary projections {} "
                    "and cannot be flipped.".format(dataset.attrs.get("name", "unknown_name"),
                                                    accepted_geos_proj_types))
        return dataset

    target_eastright, target_northup = _get_target_scene_orientation(upper_right_corner)

    area_extents_to_update = _get_dataset_area_extents_array(dataset.attrs["area"])
    current_eastright, current_northup = _get_current_scene_orientation(area_extents_to_update)

    if target_northup == current_northup and target_eastright == current_eastright:
        logger.info("Dataset {} is already in the target orientation "
                    "and will not be flipped.".format(dataset.attrs.get("name", "unknown_name")))
        return dataset

    if target_northup != current_northup:
        dataset, area_extents_to_update = _flip_dataset_data_and_area_extents(dataset, area_extents_to_update,
                                                                              "upsidedown")
    if target_eastright != current_eastright:
        dataset, area_extents_to_update = _flip_dataset_data_and_area_extents(dataset, area_extents_to_update,
                                                                              "leftright")

    dataset.attrs["area"] = _get_new_flipped_area_definition(dataset.attrs["area"], area_extents_to_update,
                                                             flip_areadef_stacking=target_northup != current_northup)

    return dataset


def _get_projection_type(dataset_area_attr):
    """Get the projection type from the crs coordinate operation method name."""
    if isinstance(dataset_area_attr, StackedAreaDefinition):
        # assumes all AreaDefinitions in a tackedAreaDefinition have the same projection
        area_crs = dataset_area_attr.defs[0].crs
    else:
        area_crs = dataset_area_attr.crs

    return area_crs.coordinate_operation.method_name


def _get_target_scene_orientation(upper_right_corner):
    """Get the target scene orientation from the target upper_right_corner.

    'NE' corresponds to target_eastright and target_northup being True.
    """
    target_northup = upper_right_corner in ["NW", "NE"]

    target_eastright = upper_right_corner in ["NE", "SE"]

    return target_eastright, target_northup


def _get_dataset_area_extents_array(dataset_area_attr):
    """Get dataset area extents in a numpy array for further flipping."""
    if isinstance(dataset_area_attr, StackedAreaDefinition):
        # array of area extents if the Area is a StackedAreaDefinition
        area_extents_to_update = np.asarray([list(area_def.area_extent) for area_def in dataset_area_attr.defs])
    else:
        # array with a single item if Area is in one piece
        area_extents_to_update = np.asarray([list(dataset_area_attr.area_extent)])
    return area_extents_to_update


def _get_current_scene_orientation(area_extents_to_update):
    """Get the current scene orientation from the area_extents."""
    # assumes all AreaDefinitions inside a StackedAreaDefinition have the same orientation
    current_northup = area_extents_to_update[0, 3] - area_extents_to_update[0, 1] > 0
    current_eastright = area_extents_to_update[0, 2] - area_extents_to_update[0, 0] > 0

    return current_eastright, current_northup


def _flip_dataset_data_and_area_extents(dataset, area_extents_to_update, flip_direction):
    """Flip the data and area extents array for a dataset."""
    logger.info("Flipping Dataset {} {}.".format(dataset.attrs.get("name", "unknown_name"), flip_direction))
    if flip_direction == "upsidedown":
        dataset = dataset[::-1, :]
        area_extents_to_update[:, [1, 3]] = area_extents_to_update[:, [3, 1]]
    elif flip_direction == "leftright":
        dataset = dataset[:, ::-1]
        area_extents_to_update[:, [0, 2]] = area_extents_to_update[:, [2, 0]]
    else:
        raise ValueError("Flip direction not recognized. Should be either 'upsidedown' or 'leftright'.")

    return dataset, area_extents_to_update


def _get_new_flipped_area_definition(dataset_area_attr, area_extents_to_update, flip_areadef_stacking):
    """Get a new area definition with updated area_extents for flipped geostationary datasets."""
    if len(area_extents_to_update) == 1:
        # just update the area extents using the AreaDefinition copy method
        new_area_def = dataset_area_attr.copy(area_extent=area_extents_to_update[0])
    else:
        # update the stacked AreaDefinitions singularly
        new_area_defs_to_stack = []
        for n_area_def, area_def in enumerate(dataset_area_attr.defs):
            new_area_defs_to_stack.append(area_def.copy(area_extent=area_extents_to_update[n_area_def]))

        # flip the order of stacking if the area is upside down
        if flip_areadef_stacking:
            new_area_defs_to_stack = new_area_defs_to_stack[::-1]

        # regenerate the StackedAreaDefinition
        new_area_def = StackedAreaDefinition(*new_area_defs_to_stack)

    return new_area_def


class GEOFlippableFileYAMLReader(FileYAMLReader):
    """Reader for flippable geostationary data."""

    def _load_dataset_with_area(self, dsid, coords, upper_right_corner="native", **kwargs):
        ds = super(GEOFlippableFileYAMLReader, self)._load_dataset_with_area(dsid, coords, **kwargs)

        if ds is not None:
            ds = _set_orientation(ds, upper_right_corner)

        return ds


class GEOSegmentYAMLReader(GEOFlippableFileYAMLReader):
    """Reader for segmented geostationary data.

    This reader pads the data to full geostationary disk if necessary.

    This reader uses an optional ``pad_data`` keyword argument that can be
    passed to :meth:`satpy.scene.Scene.load` to control if padding is done (True by
    default). Passing `pad_data=False` will return data unpadded.

    When using this class in a reader's YAML configuration, segmented file
    types (files that may have multiple segments) should specify an extra
    ``expected_segments`` piece of file_type metadata. This tells this reader
    how many total segments it should expect when padding data. Alternatively,
    the file patterns for a file type can include a ``total_segments``
    field which will be used if ``expected_segments`` is not defined. This
    will default to 1 segment.

    """

    def create_filehandlers(self, filenames, fh_kwargs=None):
        """Create file handler objects and determine expected segments for each.

        Additionally, sort the filehandlers by segment number to avoid
        issues with filenames where start_time or alphabetic sorting does not
        produce the correct order.

        """
        created_fhs = super(GEOSegmentYAMLReader, self).create_filehandlers(
            filenames, fh_kwargs=fh_kwargs)

        # add "expected_segments" information
        for fhs in created_fhs.values():
            for fh in fhs:
                # check the filename for total_segments parameter as a fallback
                ts = fh.filename_info.get("total_segments", 1)
                # if the YAML has segments explicitly specified then use that
                fh.filetype_info.setdefault("expected_segments", ts)
                # add segment key-values for FCI filehandlers
                if "segment" not in fh.filename_info:
                    fh.filename_info["segment"] = fh.filename_info.get("count_in_repeat_cycle", 1)

        self._sort_segment_filehandler_by_segment_number()
        return created_fhs

    def _sort_segment_filehandler_by_segment_number(self):
        if hasattr(self, "file_handlers"):
            for file_type in self.file_handlers.keys():
                self.file_handlers[file_type] = sorted(self.file_handlers[file_type],
                                                       key=lambda x: x.filename_info.get("segment", 0))

    def _load_dataset(self, dsid, ds_info, file_handlers, dim="y", pad_data=True):
        """Load only a piece of the dataset."""
        if not pad_data:
            return FileYAMLReader._load_dataset(dsid, ds_info,
                                                file_handlers)

        counter, expected_segments, slice_list, failure, projectable = \
            _find_missing_segments(file_handlers, ds_info, dsid)

        if projectable is None or failure:
            raise KeyError(
                "Could not load {} from any provided files".format(dsid))

        filetype = file_handlers[0].filetype_info["file_type"]
        self.empty_segment = xr.full_like(projectable, np.nan)
        for i, sli in enumerate(slice_list):
            if sli is None:
                slice_list[i] = self._get_empty_segment(dim=dim, idx=i, filetype=filetype)

        while expected_segments > counter:
            slice_list.append(self._get_empty_segment(dim=dim, idx=counter, filetype=filetype))
            counter += 1

        if dim not in slice_list[0].dims:
            return slice_list[0]
        res = xr.concat(slice_list, dim=dim)

        combined_info = file_handlers[0].combine_info(
            [p.attrs for p in slice_list])

        res.attrs = combined_info
        return res

    def _get_empty_segment(self, **kwargs):
        return self.empty_segment

    def _load_area_def(self, dsid, file_handlers, pad_data=True):
        """Load the area definition of *dsid* with padding."""
        if not pad_data:
            return _load_area_def(dsid, file_handlers)
        return self._load_area_def_with_padding(dsid, file_handlers)

    def _load_area_def_with_padding(self, dsid, file_handlers):
        """Load the area definition of *dsid* with padding."""
        # Pad missing segments between the first available and expected
        area_defs = self._pad_later_segments_area(file_handlers, dsid)

        # Add missing start segments
        area_defs = self._pad_earlier_segments_area(file_handlers, dsid, area_defs)

        # Stack the area definitions
        area_def = _stack_area_defs(area_defs)

        return area_def

    def _pad_later_segments_area(self, file_handlers, dsid):
        """Pad area definitions for missing segments that are later in sequence than the first available."""
        expected_segments = file_handlers[0].filetype_info["expected_segments"]
        filetype = file_handlers[0].filetype_info["file_type"]
        available_segments = [int(fh.filename_info.get("segment", 1)) for
                              fh in file_handlers]

        area_defs = self._get_segments_areadef_with_later_padded(file_handlers, filetype, dsid, available_segments,
                                                                 expected_segments)

        return area_defs

    def _get_segments_areadef_with_later_padded(self, file_handlers, filetype, dsid, available_segments,
                                                expected_segments):
        seg_size = None
        area_defs = {}
        for segment in range(available_segments[0], expected_segments + 1):
            try:
                idx = available_segments.index(segment)
                fh = file_handlers[idx]
                area = fh.get_area_def(dsid)
            except ValueError:
                area = self._get_new_areadef_for_padded_segment(area, filetype, seg_size, segment, padding_type="later")

            area_defs[segment] = area
            seg_size = area.shape
        return area_defs

    def _pad_earlier_segments_area(self, file_handlers, dsid, area_defs):
        """Pad area definitions for missing segments that are earlier in sequence than the first available."""
        available_segments = [int(fh.filename_info.get("segment", 1)) for
                              fh in file_handlers]
        area = file_handlers[0].get_area_def(dsid)
        seg_size = area.shape
        filetype = file_handlers[0].filetype_info["file_type"]

        for segment in range(available_segments[0] - 1, 0, -1):
            area = self._get_new_areadef_for_padded_segment(area, filetype, seg_size, segment, padding_type="earlier")
            area_defs[segment] = area
            seg_size = area.shape

        return area_defs

    def _get_new_areadef_for_padded_segment(self, area, filetype, seg_size, segment, padding_type):
        logger.debug("Padding to full disk with segment nr. %d", segment)
        new_height_px, new_ll_y, new_ur_y = self._get_y_area_extents_for_padded_segment(area, filetype, padding_type,
                                                                                        seg_size, segment)

        fill_extent = (area.area_extent[0], new_ll_y,
                       area.area_extent[2], new_ur_y)
        area = AreaDefinition("fill", "fill", "fill", area.crs,
                              seg_size[1], new_height_px,
                              fill_extent)
        return area

    def _get_y_area_extents_for_padded_segment(self, area, filetype, padding_type, seg_size, segment):
        new_height_proj_coord, new_height_px = self._get_new_areadef_heights(area, seg_size,
                                                                             segment_n=segment,
                                                                             filetype=filetype)
        if padding_type == "later":
            new_ll_y = area.area_extent[1] + new_height_proj_coord
            new_ur_y = area.area_extent[1]
        elif padding_type == "earlier":
            new_ll_y = area.area_extent[3]
            new_ur_y = area.area_extent[3] - new_height_proj_coord
        else:
            raise ValueError("Padding type not recognised.")
        return new_height_px, new_ll_y, new_ur_y

    def _get_new_areadef_heights(self, previous_area, previous_seg_size, **kwargs):
        new_height_px = previous_seg_size[0]
        new_height_proj_coord = previous_area.area_extent[1] - previous_area.area_extent[3]

        return new_height_proj_coord, new_height_px


def _stack_area_defs(area_def_dict):
    """Stack given dict of area definitions and return a StackedAreaDefinition."""
    area_defs = [area_def_dict[area_def] for
                 area_def in sorted(area_def_dict.keys())
                 if area_def is not None]

    area_def = StackedAreaDefinition(*area_defs)
    area_def = area_def.squeeze()

    return area_def


def _find_missing_segments(file_handlers, ds_info, dsid):
    """Find missing segments."""
    slice_list = []
    failure = True
    counter = 1
    expected_segments = 1

    projectable = None
    for fh in file_handlers:
        if fh.filetype_info["file_type"] in ds_info["file_type"]:
            expected_segments = fh.filetype_info["expected_segments"]

        while int(fh.filename_info.get("segment", 1)) > counter:
            slice_list.append(None)
            counter += 1
        try:
            projectable = fh.get_dataset(dsid, ds_info)
            if projectable is not None:
                slice_list.append(projectable)
                failure = False
                counter += 1
        except KeyError:
            logger.warning("Failed to load %s from %s", str(dsid), str(fh),
                           exc_info=True)

    # The last segment is missing?
    if len(slice_list) < expected_segments:
        slice_list.append(None)

    return counter, expected_segments, slice_list, failure, projectable


def _get_empty_segment_with_height(empty_segment, new_height, dim):
    """Get a new empty segment with the specified height."""
    if empty_segment.shape[0] > new_height:
        # if current empty segment is too tall, slice the DataArray
        return empty_segment[:new_height, :]
    if empty_segment.shape[0] < new_height:
        # if current empty segment is too short, pad to the new size using the empty segment values
        return empty_segment.pad(pad_width={dim : (new_height - empty_segment.shape[0], 0)},
                                 mode="constant",
                                 constant_values=empty_segment[0, 0])
    return empty_segment


class GEOVariableSegmentYAMLReader(GEOSegmentYAMLReader):
    """GEOVariableSegmentYAMLReader for handling segmented GEO products with segments of variable height.

    This YAMLReader overrides parts of the GEOSegmentYAMLReader to account for formats where the segments can
    have variable heights. It computes the sizes of the padded segments using the information available in the
    file(handlers), so that gaps of any size can be filled as needed.

    This implementation was motivated by the FCI L1c format, where the segments (called chunks in the FCI world)
    can have variable heights. It is however generic, so that any future reader can use it. The requirement
    for the reader is to have a method called `get_segment_position_info`, returning a dictionary containing
    the positioning info for each segment (see example in
    :func:`satpy.readers.fci_l1c_nc.FCIL1cNCFileHandler.get_segment_position_info`).

    For more information on please see the documentation of :func:`satpy.readers.core.yaml_reader.GEOSegmentYAMLReader`.
    """

    def __init__(self,
                 config_dict,
                 filter_parameters=None,
                 filter_filenames=True,
                 **kwargs):
        """Initialise the GEOVariableSegmentYAMLReader object."""
        super().__init__(config_dict, filter_parameters, filter_filenames, **kwargs)
        self.segment_heights = cache(self._segment_heights)
        self.segment_infos = dict()

    def _extract_segment_location_dicts(self, filetype):
        self._initialise_segment_infos(filetype)
        self._collect_segment_position_infos(filetype)
        return

    def _collect_segment_position_infos(self, filetype):
        # collect the segment positioning infos for all available segments
        for fh in self.file_handlers[filetype]:
            chk_infos = fh.get_segment_position_info()
            chk_infos.update({"segment_nr": fh.filename_info["segment"] - 1})
            self.segment_infos[filetype]["available_segment_infos"].append(chk_infos)

    def _initialise_segment_infos(self, filetype):
        # initialise the segment info for this filetype
        filetype_fhs_sample = self.file_handlers[filetype][0]
        exp_segment_nr = filetype_fhs_sample.filetype_info["expected_segments"]
        grid_width_to_grid_type = _get_grid_width_to_grid_type(filetype_fhs_sample.get_segment_position_info())
        self.segment_infos.update({filetype: {"available_segment_infos": [],
                                              "expected_segments": exp_segment_nr,
                                              "grid_width_to_grid_type": grid_width_to_grid_type}})

    def _get_empty_segment(self, dim=None, idx=None, filetype=None):
        grid_width = self.empty_segment.shape[1]
        segment_height = self.segment_heights(filetype, grid_width)[idx]
        return _get_empty_segment_with_height(self.empty_segment, segment_height, dim=dim)

    def _segment_heights(self, filetype, grid_width):
        """Compute optimal padded segment heights (in number of pixels) based on the location of available segments."""
        self._extract_segment_location_dicts(filetype)
        grid_type = self.segment_infos[filetype]["grid_width_to_grid_type"][grid_width]
        segment_heights = _compute_optimal_missing_segment_heights(self.segment_infos[filetype], grid_type, grid_width)
        return segment_heights

    def _get_new_areadef_heights(self, previous_area, previous_seg_size, segment_n=None, filetype=None):
        # retrieve the segment height in number of pixels
        grid_width = previous_seg_size[1]
        new_height_px = self.segment_heights(filetype, grid_width)[segment_n - 1]
        # scale the previous vertical area extent using the new pixel height
        prev_area_extent = previous_area.area_extent[1] - previous_area.area_extent[3]
        new_height_proj_coord = prev_area_extent * new_height_px / previous_seg_size[0]

        return new_height_proj_coord, new_height_px


def _get_grid_width_to_grid_type(seg_info):
    grid_width_to_grid_type = dict()
    for grid_type, grid_type_seg_info in seg_info.items():
        grid_width_to_grid_type.update({grid_type_seg_info["grid_width"]: grid_type})
    return grid_width_to_grid_type


def _compute_optimal_missing_segment_heights(seg_infos, grid_type, expected_vertical_size):
    # initialise positioning arrays
    segment_start_rows, segment_end_rows, segment_heights = _init_positioning_arrays_for_variable_padding(
        seg_infos["available_segment_infos"], grid_type, seg_infos["expected_segments"])

    # populate start row of first segment and end row of last segment with known values
    segment_start_rows[0] = 1
    segment_end_rows[seg_infos["expected_segments"] - 1] = expected_vertical_size

    # find missing segments and group contiguous missing segments together
    missing_segments = np.where(segment_heights == 0)[0]
    groups_missing_segments = np.split(missing_segments, np.where(np.diff(missing_segments) > 1)[0] + 1)

    for group in groups_missing_segments:
        _compute_positioning_data_for_missing_group(segment_start_rows, segment_end_rows, segment_heights, group)

    return segment_heights.astype("int")


def _compute_positioning_data_for_missing_group(segment_start_rows, segment_end_rows, segment_heights, group):
    _populate_group_start_end_row_using_neighbour_segments(group, segment_end_rows, segment_start_rows)
    proposed_sizes_missing_segments = _compute_proposed_sizes_of_missing_segments_in_group(group, segment_end_rows,
                                                                                           segment_start_rows)
    _populate_start_end_rows_of_missing_segments_with_proposed_sizes(group, proposed_sizes_missing_segments,
                                                                     segment_start_rows, segment_end_rows,
                                                                     segment_heights)


def _populate_start_end_rows_of_missing_segments_with_proposed_sizes(group, proposed_sizes_missing_segments,
                                                                     segment_start_rows, segment_end_rows,
                                                                     segment_heights):
    for n in range(len(group)):
        # start of first and end of last missing segment have been populated already
        if n != 0:
            segment_start_rows[group[n]] = segment_start_rows[group[n - 1]] + proposed_sizes_missing_segments[n] + 1
        if n != len(group) - 1:
            segment_end_rows[group[n]] = segment_start_rows[group[n]] + proposed_sizes_missing_segments[n]
        segment_heights[group[n]] = proposed_sizes_missing_segments[n]


def _compute_proposed_sizes_of_missing_segments_in_group(group, segment_end_rows, segment_start_rows):
    size_group_gap = segment_end_rows[group[-1]] - segment_start_rows[group[0]] + 1
    proposed_sizes_missing_segments = split_integer_in_most_equal_parts(size_group_gap, len(group))
    return proposed_sizes_missing_segments


def _populate_group_start_end_row_using_neighbour_segments(group, segment_end_rows, segment_start_rows):
    # if group is at the start/end of the full-disk, we know the start/end value already
    if segment_start_rows[group[0]] == 0:
        _populate_group_start_row_using_previous_segment(group, segment_end_rows, segment_start_rows)
    if segment_end_rows[group[-1]] == 0:
        _populate_group_end_row_using_later_segment(group, segment_end_rows, segment_start_rows)


def _populate_group_end_row_using_later_segment(group, segment_end_rows, segment_start_rows):
    segment_end_rows[group[-1]] = segment_start_rows[group[-1] + 1] - 1


def _populate_group_start_row_using_previous_segment(group, segment_end_rows, segment_start_rows):
    segment_start_rows[group[0]] = segment_end_rows[group[0] - 1] + 1


def _init_positioning_arrays_for_variable_padding(chk_infos, grid_type, exp_segment_nr):
    segment_heights = np.zeros(exp_segment_nr)
    segment_start_rows = np.zeros(exp_segment_nr)
    segment_end_rows = np.zeros(exp_segment_nr)

    _populate_positioning_arrays_with_available_segment_info(chk_infos, grid_type, segment_start_rows, segment_end_rows,
                                                             segment_heights)
    return segment_start_rows, segment_end_rows, segment_heights


def _populate_positioning_arrays_with_available_segment_info(chk_infos, grid_type, segment_start_rows, segment_end_rows,
                                                             segment_heights):
    for chk_info in chk_infos:
        current_fh_segment_nr = chk_info["segment_nr"]
        segment_heights[current_fh_segment_nr] = chk_info[grid_type]["segment_height"]
        segment_start_rows[current_fh_segment_nr] = chk_info[grid_type]["start_position_row"]
        segment_end_rows[current_fh_segment_nr] = chk_info[grid_type]["end_position_row"]


def split_integer_in_most_equal_parts(x, n):
    """Split an integer number x in n parts that are as equally-sizes as possible."""
    if x % n == 0:
        return np.repeat(x // n, n).astype("int")
    else:
        # split the remainder amount over the last remainder parts
        remainder = int(x % n)
        mod = int(x // n)
        ar = np.repeat(mod, n)
        ar[-remainder:] = mod + 1
        return ar.astype("int")



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositors."""

import dask
import dask.array as da
import numpy as np
import xarray as xr

from satpy.tests.utils import CustomScheduler


class TestRealisticColors:
    """Test the SEVIRI Realistic Colors compositor."""

    def test_realistic_colors(self):
        """Test the compositor."""
        from satpy.composites.seviri import RealisticColors

        vis06 = xr.DataArray(da.arange(0, 15, dtype=np.float32).reshape(3, 5), dims=("y", "x"),
                             attrs={"foo": "foo"})
        vis08 = xr.DataArray(da.arange(15, 0, -1, dtype=np.float32).reshape(3, 5), dims=("y", "x"),
                             attrs={"bar": "bar"})
        hrv = xr.DataArray(6 * da.ones((3, 5), dtype=np.float32), dims=("y", "x"),
                           attrs={"baz": "baz"})

        expected_red = np.array([[0.0, 2.733333, 4.9333334, 6.6, 7.733333],
                                 [8.333333, 8.400001, 7.9333334, 7.0, 6.0],
                                 [5.0, 4.0, 3.0, 2.0, 1.0]], dtype=np.float32)
        expected_green = np.array([
            [15.0, 12.266666, 10.066668, 8.400001, 7.2666664],
            [6.6666665, 6.6000004, 7.0666666, 8.0, 9.0],
            [10.0, 11.0, 12.0, 13.0, 14.0]], dtype=np.float32)

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = RealisticColors("Ni!")
            res = comp((vis06, vis08, hrv))

        arr = res.values

        assert res.dtype == np.float32
        np.testing.assert_allclose(arr[0, :, :], expected_red)
        np.testing.assert_allclose(arr[1, :, :], expected_green)
        np.testing.assert_allclose(arr[2, :, :], 3.0)



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Writers subpackage."""
from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

IMPORT_PATHS = {
    "Writer": "satpy.writers.core.base",
    "ImageWriter": "satpy.writers.core.image",
    "add_overlay": "satpy.enhancements.overlays",
    "add_decorate": "satpy.enhancements.overlays",
    "add_scale": "satpy.enhancements.overlays",
    "add_logo": "satpy.enhancements.overlays",
    "add_text": "satpy.enhancements.overlays",
    "split_results": "satpy.writers.core.compute",
    "group_results_by_output_file": "satpy.writers.core.compute",
    "compute_writer_results": "satpy.writers.core.compute",
    "get_enhanced_image": "satpy.enhancements.enhancer",
    "read_writer_config": "satpy.writers.core.config",
    "load_writer_configs": "satpy.writers.core.config",
    "load_writer": "satpy.writers.core.config",
    "configs_for_writer": "satpy.writers.core.config",
    "available_writers": "satpy.writers.core.config",
}

def __getattr__(name: str) -> Any:
    if name == "show":
        raise AttributeError("The 'show' function has been removed. Use 'get_enhanced_image(data_arr).show()' instead.")
    if name == "to_image":
        raise AttributeError("The 'to_image' function has been removed. Use 'trollimage.xrimage.XRImage' instead.")

    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module {__name__} has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Functions and utilities for downloading ancillary data."""

import logging
import os

import satpy

logger = logging.getLogger(__name__)

_FILE_REGISTRY = {}
_FILE_URLS = {}
RUNNING_TESTS = False


def register_file(url, filename, component_type=None, known_hash=None):
    """Register file for future retrieval.

    This function only prepares Satpy to be able to download and cache the
    provided file. It will not download the file. See
    :func:`satpy.aux_download.retrieve` for more information.

    Args:
        url (str): URL where remote file can be downloaded.
        filename (str): Filename used to identify and store the downloaded
            file as.
        component_type (str or None): Name of the type of Satpy component that
            will use this file. Typically "readers", "composites", "writers",
            or "enhancements" for consistency. This will be prepended to the
            filename when storing the data in the cache.
        known_hash (str): Hash used to verify the file is downloaded correctly.
            See https://www.fatiando.org/pooch/v1.3.0/beginner.html#hashes
            for more information. If not provided then the file is not checked.

    Returns:
        Cache key that can be used to retrieve the file later. The cache key
        consists of the ``component_type`` and provided ``filename``. This
        should be passed to :func:`satpy.aux_download.retrieve` when the
        file will be used.

    """
    fname = _generate_filename(filename, component_type)
    _FILE_REGISTRY[fname] = known_hash
    _FILE_URLS[fname] = url
    return fname


def _generate_filename(filename, component_type):
    if filename is None:
        return None
    path = filename
    if component_type:
        path = "/".join([component_type, path])
    return path


def _retrieve_offline(data_dir, cache_key):
    logger.debug("Downloading auxiliary files is turned off, will check "
                 "local files.")
    local_file = os.path.join(data_dir, *cache_key.split("/"))
    if not os.path.isfile(local_file):
        raise RuntimeError("Satpy 'download_aux' setting is False meaning "
                           "no new files will be downloaded and the local "
                           "file '{}' does not exist.".format(local_file))
    return local_file


def _should_download(cache_key):
    """Check if we're running tests and can download this file."""
    return not RUNNING_TESTS or "README" in cache_key


def retrieve(cache_key, pooch_kwargs=None):
    """Download and cache the file associated with the provided ``cache_key``.

    Cache location is controlled by the config ``data_dir`` key. See
    :ref:`data_dir_setting` for more information.

    Args:
        cache_key (str): Cache key returned by
            :func:`~satpy.aux_download.register_file`.
        pooch_kwargs (dict or None): Extra keyword arguments to pass to
            :meth:`pooch.Pooch.fetch`.

    Returns:
        Local path of the cached file.


    """
    import pooch

    pooch_kwargs = pooch_kwargs or {}

    path = satpy.config.get("data_dir")
    if not satpy.config.get("download_aux"):
        return _retrieve_offline(path, cache_key)
    if not _should_download(cache_key):
        raise RuntimeError("Auxiliary data download is not allowed during "
                           "tests. Mock the appropriate components of your "
                           "tests to not need the 'retrieve' function.")
    # reuse data directory as the default URL where files can be downloaded from
    pooch_obj = pooch.create(path, path, registry=_FILE_REGISTRY,
                             urls=_FILE_URLS)
    return pooch_obj.fetch(cache_key, **pooch_kwargs)


def _retrieve_all_with_pooch(pooch_kwargs):
    import pooch

    if pooch_kwargs is None:
        pooch_kwargs = {}
    path = satpy.config.get("data_dir")
    pooch_obj = pooch.create(path, path, registry=_FILE_REGISTRY,
                             urls=_FILE_URLS)
    for fname in _FILE_REGISTRY:
        logger.info("Downloading extra data file '%s'...", fname)
        pooch_obj.fetch(fname, **pooch_kwargs)


def retrieve_all(readers=None, writers=None, composite_sensors=None,
                 pooch_kwargs=None):
    """Find cache-able data files for Satpy and download them.

    The typical use case for this function is to download all ancillary files
    before going to an environment/system that does not have internet access.

    Args:
        readers (list or None): Limit searching to these readers. If not
            specified or ``None`` then all readers are searched. If an
            empty list then no readers are searched.
        writers (list or None): Limit searching to these writers. If not
            specified or ``None`` then all writers are searched. If an
            empty list then no writers are searched.
        composite_sensors (list or None): Limit searching to composite
            configuration files for these sensors. If ``None`` then all sensor
            configs will be searched. If an empty list then no composites
            will be searched.
        pooch_kwargs (dict): Additional keyword arguments to pass to pooch
            ``fetch``.

    """
    if not satpy.config.get("download_aux"):
        raise RuntimeError("Satpy 'download_aux' setting is False so no files "
                           "will be downloaded.")

    find_registerable_files(readers=readers,
                            writers=writers,
                            composite_sensors=composite_sensors)
    _retrieve_all_with_pooch(pooch_kwargs)
    logger.info("Done downloading all extra files.")


def find_registerable_files(readers=None, writers=None,
                            composite_sensors=None):
    """Load all Satpy components so they can be downloaded.

    Args:
        readers (list or None): Limit searching to these readers. If not
            specified or ``None`` then all readers are searched. If an
            empty list then no readers are searched.
        writers (list or None): Limit searching to these writers. If not
            specified or ``None`` then all writers are searched. If an
            empty list then no writers are searched.
        composite_sensors (list or None): Limit searching to composite
            configuration files for these sensors. If ``None`` then all sensor
            configs will be searched. If an empty list then no composites
            will be searched.

    """
    _find_registerable_files_compositors(composite_sensors)
    _find_registerable_files_readers(readers)
    _find_registerable_files_writers(writers)
    return sorted(_FILE_REGISTRY.keys())


def _find_registerable_files_compositors(sensors=None):
    """Load all compositor configs so that files are registered.

    Compositor objects should register files when they are initialized.

    """
    from satpy.composites.config_loader import all_composite_sensors, load_compositor_configs_for_sensors
    if sensors is None:
        sensors = all_composite_sensors()
    if sensors:
        mods = load_compositor_configs_for_sensors(sensors)[1]
        _register_modifier_files(mods)


def _register_modifier_files(modifiers):
    for mod_sensor_dict in modifiers.values():
        for mod_name, (mod_cls, mod_props) in mod_sensor_dict.items():
            try:
                mod_cls(**mod_props)
            except (ValueError, RuntimeError):
                logger.error("Could not initialize modifier '%s' for "
                             "auxiliary download registration.", mod_name)


def _find_registerable_files_readers(readers=None):
    """Load all readers so that files are registered."""
    import yaml

    from satpy.readers.core.config import configs_for_reader
    from satpy.readers.core.loading import load_reader
    for reader_configs in configs_for_reader(reader=readers):
        try:
            load_reader(reader_configs)
        except (ModuleNotFoundError, yaml.YAMLError):
            continue


def _find_registerable_files_writers(writers=None):
    """Load all writers so that files are registered."""
    from satpy.writers.core.config import configs_for_writer, load_writer_configs

    for writer_configs in configs_for_writer(writer=writers):
        try:
            load_writer_configs(writer_configs)
        except ValueError:
            continue


class DataDownloadMixin:
    """Mixin class for Satpy components to download files.

    This class simplifies the logic needed to download and cache data files
    needed for operations in a Satpy component (readers, writers, etc). It
    does this in a two step process where files that might be downloaded are
    "registered" and then "retrieved" when they need to be used.

    To use this class include it as one of the subclasses of your Satpy
    component. Then in the ``__init__`` method, call the
    ``register_data_files`` function during initialization.

    .. note::

        This class is already included in the ``FileYAMLReader`` and
        ``Writer`` base classes. There is no need to define a custom
        class.

    The below code is shown as an example::

        from satpy.readers.core.yaml_reader import AbstractYAMLReader
        from satpy.aux_download import DataDownloadMixin

        class MyReader(AbstractYAMLReader, DataDownloadMixin):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.register_data_files()

    This class expects data files to be configured in either a
    ``self.info['data_files']`` (standard for readers/writers) or
    ``self.config['data_files']`` list. The ``data_files`` item
    itself is a list of dictionaries. This information can also be
    passed directly to ``register_data_files`` for more complex cases.
    In YAML, for a reader, this might look like this::

        reader:
            name: abi_l1b
            short_name: ABI L1b
            long_name: GOES-R ABI Level 1b
            ... other metadata ...
            data_files:
              - url: "https://example.com/my_data_file.dat"
              - url: "https://raw.githubusercontent.com/pytroll/satpy/main/README.rst"
                known_hash: "sha256:5891286b63e7745de08c4b0ac204ad44cfdb9ab770309debaba90308305fa759"
              - url: "https://raw.githubusercontent.com/pytroll/satpy/main/RELEASING.md"
                filename: "satpy_releasing.md"

    In this example we register two files that might be downloaded.
    If ``known_hash`` is not provided or None (null in YAML) then the data
    file will not be checked for validity when downloaded. See
    :func:`~satpy.aux_download.register_file` for more information. You can
    optionally specify ``filename`` to define the in-cache name when this file
    is downloaded. This can be useful in cases when the filename can not be
    easily determined from the URL.

    When it comes time to needing the file, you can retrieve the local path
    by calling ``~satpy.aux_download.retrieve(cache_key)`` with the
    "cache key" generated during registration. These keys will be in the
    format: ``<component_type>/<filename>``. For a
    reader this would be ``readers/satpy_release.md``.

    This Mixin is not the only way to register and download files for a
    Satpy component, but is the most generic and flexible. Feel free to
    use the :func:`~satpy.aux_download.register_file` and
    :func:`~satpy.aux_download.retrieve` functions directly.
    However, :meth:`~satpy.aux_download.find_registerable_files` must also
    be updated to support your component (if files are not register during
    initialization).

    """

    DATA_FILE_COMPONENTS = {
        "reader": "readers",
        "writer": "writers",
        "composit": "composites",
        "modifi": "modifiers",
        "corr": "modifiers",
    }

    @property
    def _data_file_component_type(self):
        cls_name = self.__class__.__name__.lower()
        for cls_name_sub, comp_type in self.DATA_FILE_COMPONENTS.items():
            if cls_name_sub in cls_name:
                return comp_type
        return "other"

    def register_data_files(self, data_files=None):
        """Register a series of files that may be downloaded later.

        See :class:`~satpy.aux_download.DataDownloadMixin` for more
        information on the assumptions and structure of the data file
        configuration dictionary.

        """
        comp_type = self._data_file_component_type
        if data_files is None:
            df_parent = getattr(self, "info", self.config)
            data_files = df_parent.get("data_files", [])
        cache_keys = []
        for data_file_entry in data_files:
            cache_key = self._register_data_file(data_file_entry, comp_type)
            cache_keys.append(cache_key)
        return cache_keys

    @staticmethod
    def _register_data_file(data_file_entry, comp_type):
        url = data_file_entry["url"]
        filename = data_file_entry.get("filename", os.path.basename(url))
        known_hash = data_file_entry.get("known_hash")
        return register_file(url, filename, component_type=comp_type,
                             known_hash=known_hash)


def retrieve_all_cmd(argv=None):
    """Call 'retrieve_all' function from console script 'satpy_retrieve_all'."""
    import argparse
    parser = argparse.ArgumentParser(description="Download auxiliary data files used by Satpy.")
    parser.add_argument("--data-dir",
                        help="Override 'SATPY_DATA_DIR' for destination of "
                             "downloaded files. This does NOT change the "
                             "directory Satpy will look at when searching "
                             "for files outside of this script.")
    parser.add_argument("--composite-sensors", nargs="*",
                        help="Limit loaded composites for the specified "
                             "sensors. If specified with no arguments, "
                             "no composite files will be downloaded.")
    parser.add_argument("--readers", nargs="*",
                        help="Limit searching to these readers. If specified "
                             "with no arguments, no reader files will be "
                             "downloaded.")
    parser.add_argument("--writers", nargs="*",
                        help="Limit searching to these writers. If specified "
                             "with no arguments, no writer files will be "
                             "downloaded.")
    args = parser.parse_args(argv)

    logging.basicConfig(level=logging.INFO)

    if args.data_dir is None:
        args.data_dir = satpy.config.get("data_dir")

    with satpy.config.set(data_dir=args.data_dir):
        retrieve_all(readers=args.readers, writers=args.writers,
                     composite_sensors=args.composite_sensors)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for SAR compositors."""

import unittest


class TestSARComposites(unittest.TestCase):
    """Test SAR-specific composites."""

    def test_sar_ice(self):
        """Test creating a the sar_ice composite."""
        import dask.array as da
        import numpy as np
        import xarray as xr

        from satpy.composites.sar import SARIce

        rows = 2
        cols = 2
        comp = SARIce("sar_ice", prerequisites=("hh", "hv"),
                      standard_name="sar-ice")
        hh = xr.DataArray(da.zeros((rows, cols), chunks=25) + 2000,
                          dims=("y", "x"),
                          attrs={"name": "hh"})
        hv = xr.DataArray(da.zeros((rows, cols), chunks=25) + 1000,
                          dims=("y", "x"),
                          attrs={"name": "hv"})

        res = comp((hh, hv))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "sar_ice"
        assert res.attrs["standard_name"] == "sar-ice"
        data = res.compute()
        np.testing.assert_allclose(data.sel(bands="R"), 31.58280822)
        np.testing.assert_allclose(data.sel(bands="G"), 159869.56789876)
        np.testing.assert_allclose(data.sel(bands="B"), 44.68138191)

    def test_sar_ice_log(self):
        """Test creating a the sar_ice_log composite."""
        import dask.array as da
        import numpy as np
        import xarray as xr

        from satpy.composites.sar import SARIceLog

        rows = 2
        cols = 2
        comp = SARIceLog("sar_ice_log", prerequisites=("hh", "hv"),
                         standard_name="sar-ice-log")
        hh = xr.DataArray(da.zeros((rows, cols), chunks=25) - 10,
                          dims=("y", "x"),
                          attrs={"name": "hh"})
        hv = xr.DataArray(da.zeros((rows, cols), chunks=25) - 20,
                          dims=("y", "x"),
                          attrs={"name": "hv"})

        res = comp((hh, hv))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "sar_ice_log"
        assert res.attrs["standard_name"] == "sar-ice-log"
        data = res.compute()
        np.testing.assert_allclose(data.sel(bands="R"), -20)
        np.testing.assert_allclose(data.sel(bands="G"), -4.6)
        np.testing.assert_allclose(data.sel(bands="B"), -10)



"""Module for testing the satpy.readers.viirs_l2 module."""

import datetime as dt
import os
from unittest import mock

import numpy as np
import pytest

from satpy.readers.core.loading import load_reader
from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.float32
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(
    DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1], dtype=DEFAULT_FILE_DTYPE
).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeNetCDF4FileHandlerVIIRSL2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        date = filename_info.get("start_time", dt.datetime(2023, 12, 30, 22, 30, 0))
        file_type = filename[:6]
        num_lines = DEFAULT_FILE_SHAPE[0]
        num_pixels = DEFAULT_FILE_SHAPE[1]
        num_scans = 5
        file_content = {
            "/dimension/number_of_scans": num_scans,
            "/dimension/number_of_lines": num_lines,
            "/dimension/number_of_pixels": num_pixels,
            "/attr/time_coverage_start": date.strftime("%Y-%m-%dT%H:%M:%S.000Z"),
            "/attr/time_coverage_end": (date + dt.timedelta(minutes=6)).strftime(
                "%Y-%m-%dT%H:%M:%S.000Z"
            ),
            "/attr/orbit_number": 26384,
            "/attr/instrument": "VIIRS",
            "/attr/platform": "Suomi-NPP",
        }
        self._fill_contents_with_default_data(file_content, file_type)
        convert_file_content_to_data_array(file_content)
        return file_content

    def _fill_contents_with_default_data(self, file_content, file_type):
        """Fill file contents with default data."""
        if file_type.startswith("CLD"):
            file_content["geolocation_data/latitude"] = DEFAULT_LAT_DATA
            file_content["geolocation_data/longitude"] = DEFAULT_LON_DATA
            if file_type == "CLDPRO":
                file_content["geophysical_data/Cloud_Top_Height"] = DEFAULT_FILE_DATA
            elif file_type == "CLDMSK":
                file_content[
                    "geophysical_data/Clear_Sky_Confidence"
                ] = DEFAULT_FILE_DATA
        elif file_type == "AERDB_":
            file_content["Latitude"] = DEFAULT_LAT_DATA
            file_content["Longitude"] = DEFAULT_LON_DATA
            file_content["Angstrom_Exponent_Land_Ocean_Best_Estimate"] = DEFAULT_FILE_DATA
            file_content["Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate"] = DEFAULT_FILE_DATA


class TestVIIRSL2FileHandler:
    """Test VIIRS_L2 Reader."""
    yaml_file = "viirs_l2.yaml"

    def setup_method(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_l2 import VIIRSL2FileHandler

        self.reader_configs = config_search_paths(
            os.path.join("readers", self.yaml_file)
        )
        self.p = mock.patch.object(
            VIIRSL2FileHandler, "__bases__", (FakeNetCDF4FileHandlerVIIRSL2,)
        )
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    @pytest.mark.parametrize(
        "filename",
        [
            ("CLDPROP_L2_VIIRS_SNPP.A2023364.2230.011.2023365115856.nc"),
            ("CLDMSK_L2_VIIRS_SNPP.A2023364.2230.001.2023365105952.nc"),
            ("AERDB_L2_VIIRS_SNPP.A2023364.2230.011.2023365113427.nc"),
        ],
    )
    def test_init(self, filename):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader

        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([filename])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    @pytest.mark.parametrize(
        ("filename", "datasets"),
        [
            pytest.param("CLDPROP_L2_VIIRS_SNPP.A2023364.2230.011.2023365115856.nc",
                         ["Cloud_Top_Height"], id="CLDPROP"),
            pytest.param("CLDMSK_L2_VIIRS_SNPP.A2023364.2230.001.2023365105952.nc",
                         ["Clear_Sky_Confidence"], id="CLDMSK"),
            pytest.param("AERDB_L2_VIIRS_SNPP.A2023364.2230.011.2023365113427.nc",
                         ["Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate",
                          "Angstrom_Exponent_Land_Ocean_Best_Estimate"], id="AERDB"),
        ],
    )
    def test_load_l2_files(self, filename, datasets):
        """Test L2 File Loading."""
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([filename])
        r.create_filehandlers(loadables)
        loaded_datasets = r.load(datasets)
        assert len(loaded_datasets) == len(datasets)
        for d in loaded_datasets.values():
            assert d.shape == DEFAULT_FILE_SHAPE
            assert d.dims == ("y", "x")
            assert d.attrs["sensor"] == "viirs"
            d_np = d.compute()
            assert d.dtype == d_np.dtype
            assert d.dtype == np.float32



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ahi_l1b_gridded_bin reader tests package."""

import os
import shutil
import tempfile
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
from pyresample.geometry import AreaDefinition

from satpy.readers.ahi_l1b_gridded_bin import AHI_LUT_NAMES, AHIGriddedFileHandler


class TestAHIGriddedArea(unittest.TestCase):
    """Test the AHI gridded reader definition."""

    def setUp(self):
        """Create fake data for testing."""
        self.FULLDISK_SIZES = {0.005: {"x_size": 24000,
                                       "y_size": 24000},
                               0.01: {"x_size": 12000,
                                      "y_size": 12000},
                               0.02: {"x_size": 6000,
                                      "y_size": 6000}}

        self.AHI_FULLDISK_EXTENT = [85., -60., 205., 60.]

    @staticmethod
    def make_fh(filetype, area="fld"):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            fh = AHIGriddedFileHandler("somefile",
                                       {"area": area},
                                       filetype_info={"file_type": filetype})
            return fh

    def test_low_res(self):
        """Check size of the low resolution (2km) grid."""
        tmp_fh = self.make_fh("tir.01")
        assert self.FULLDISK_SIZES[0.02]["x_size"] == tmp_fh.ncols
        assert self.FULLDISK_SIZES[0.02]["y_size"] == tmp_fh.nlines

    def test_med_res(self):
        """Check size of the low resolution (1km) grid."""
        tmp_fh = self.make_fh("vis.02")
        assert self.FULLDISK_SIZES[0.01]["x_size"] == tmp_fh.ncols
        assert self.FULLDISK_SIZES[0.01]["y_size"] == tmp_fh.nlines

    def test_hi_res(self):
        """Check size of the low resolution (0.5km) grid."""
        tmp_fh = self.make_fh("ext.01")
        assert self.FULLDISK_SIZES[0.005]["x_size"] == tmp_fh.ncols
        assert self.FULLDISK_SIZES[0.005]["y_size"] == tmp_fh.nlines

    def test_area_def(self):
        """Check that a valid full disk area is produced."""
        good_area = AreaDefinition("gridded_himawari",
                                   "A gridded Himawari area",
                                   "longlat",
                                   "EPSG:4326",
                                   self.FULLDISK_SIZES[0.01]["x_size"],
                                   self.FULLDISK_SIZES[0.01]["y_size"],
                                   self.AHI_FULLDISK_EXTENT)

        tmp_fh = self.make_fh("vis.01")
        tmp_fh.get_area_def(None)
        assert tmp_fh.area == good_area

    def test_bad_area(self):
        """Ensure an error is raised for an usupported area."""
        tmp_fh = self.make_fh("ext.01")
        tmp_fh.areaname = "scanning"
        with pytest.raises(NotImplementedError):
            tmp_fh.get_area_def(None)
        with pytest.raises(NotImplementedError):
            self.make_fh("ext.01", area="scanning")


class TestAHIGriddedFileCalibration(unittest.TestCase):
    """Test case for the file calibration types."""

    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            in_fname = "test_file"
            fh = AHIGriddedFileHandler(in_fname,
                                       {"area": "fld"},
                                       filetype_info={"file_type": "tir.01"})
            self.fh = fh

    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._get_luts")
    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.os.path.exists")
    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.np.loadtxt")
    def test_calibrate(self, np_loadtxt, os_exist, get_luts):
        """Test the calibration modes of AHI using the LUTs."""
        load_return = np.squeeze(np.dstack([np.arange(0, 2048, 1),
                                            np.arange(0, 120, 0.05859375)]))

        np_loadtxt.return_value = load_return
        get_luts.return_value = True

        in_data = np.array([[100., 300., 500.],
                           [800., 1500., 2040.]])

        refl_out = np.array([[5.859375, 17.578125, 29.296875],
                             [46.875, 87.890625, 119.53125]])

        os_exist.return_value = False
        # Check that the LUT download is called if we don't have the LUTS
        self.fh.calibrate(in_data, "reflectance")
        get_luts.assert_called()

        os_exist.return_value = True
        # Ensure results equal if no calibration applied
        out_data = self.fh.calibrate(in_data, "counts")
        np.testing.assert_equal(in_data, out_data)

        # Now ensure results equal if LUT calibration applied
        out_data = self.fh.calibrate(in_data, "reflectance")
        np.testing.assert_allclose(refl_out, out_data)

        # Check that exception is raised if bad calibration is passed
        with pytest.raises(NotImplementedError):
            self.fh.calibrate(in_data, "lasers")

        # Check that exception is raised if no file is present
        np_loadtxt.side_effect = FileNotFoundError
        with pytest.raises(FileNotFoundError):
            self.fh.calibrate(in_data, "reflectance")


class TestAHIGriddedFileHandler(unittest.TestCase):
    """Test case for the file reading."""

    def new_unzip(fname):
        """Fake unzipping."""
        if fname[-3:] == "bz2":
            return fname[:-4]

    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.unzip_file",
                mock.MagicMock(side_effect=new_unzip))
    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            in_fname = "test_file.bz2"
            fh = AHIGriddedFileHandler(in_fname,
                                       {"area": "fld"},
                                       filetype_info={"file_type": "tir.01"})

            # Check that the filename is altered for bz2 format files
            assert in_fname != fh.filename
            self.fh = fh

        key = {"calibration": "counts",
               "name": "vis.01"}
        info = {"units": "unitless",
                "standard_name": "vis.01",
                "wavelength": 10.8,
                "resolution": 0.05}
        self.key = key
        self.info = info

    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.np.memmap")
    def test_dataread(self, memmap):
        """Check that a dask array is returned from the read function."""
        test_arr = np.zeros((10, 10))
        memmap.return_value = test_arr
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            res = self.fh._read_data(mock.MagicMock())
            np.testing.assert_allclose(res, da.from_array(test_arr))

    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._read_data")
    def test_get_dataset(self, mocked_read):
        """Check that a good dataset is returned on request."""
        m = mock.mock_open()

        out_data = np.array([[100., 300., 500.],
                             [800., 1500., 2040.]])
        mocked_read.return_value = out_data
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            res = self.fh.get_dataset(self.key, self.info)
            mocked_read.assert_called()
            # Check output data is correct
            np.testing.assert_allclose(res.values, out_data)
            # Also check a couple of attributes
            assert res.attrs["name"] == self.key["name"]
            assert res.attrs["wavelength"] == self.info["wavelength"]

    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("os.remove")
    def test_destructor(self, exist_patch, remove_patch):
        """Check that file handler deletes files if needed."""
        del self.fh
        remove_patch.assert_called()


class TestAHIGriddedLUTs(unittest.TestCase):
    """Test case for the downloading and preparing LUTs."""

    def mocked_ftp_dl(fname):
        """Fake download of LUT tar file by creating a local tar."""
        import os
        import tarfile
        import tempfile

        with tarfile.open(fname, "w:gz") as tar_handle:
            for namer in AHI_LUT_NAMES:
                tmpf = os.path.join(tempfile.tempdir, namer)
                with open(tmpf, "w") as tmp_fid:
                    tmp_fid.write("TEST\n")
                tar_handle.add(tmpf, arcname="count2tbb_v102/"+namer)
                os.remove(tmpf)

    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            in_fname = "test_file"
            fh = AHIGriddedFileHandler(in_fname,
                                       {"area": "fld"},
                                       filetype_info={"file_type": "tir.01"})
            self.fh = fh

        key = {"calibration": "counts",
               "name": "vis.01"}
        info = {"units": "unitless",
                "standard_name": "vis.01",
                "wavelength": 10.8,
                "resolution": 0.05}
        self.key = key
        self.info = info

    def tearDown(self):
        """Remove files and directories created by the tests."""
        for lut_name in AHI_LUT_NAMES:
            tmp_filename = os.path.join(self.fh.lut_dir, lut_name)
            if os.path.isfile(tmp_filename):
                os.remove(tmp_filename)
        if os.path.isdir(self.fh.lut_dir):
            shutil.rmtree(self.fh.lut_dir)

    @mock.patch("satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._download_luts",
                mock.MagicMock(side_effect=mocked_ftp_dl))
    def test_get_luts(self):
        """Check that the function to download LUTs operates successfully."""
        tempdir = tempfile.gettempdir()
        self.fh._get_luts()
        assert not os.path.exists(os.path.join(tempdir, "count2tbb_v102/"))
        for lut_name in AHI_LUT_NAMES:
            assert os.path.isfile(os.path.join(self.fh.lut_dir, lut_name))

    @mock.patch("urllib.request.urlopen")
    @mock.patch("shutil.copyfileobj")
    def test_download_luts(self, mock_dl, mock_shutil):
        """Test that the FTP library is called for downloading LUTS."""
        m = mock.mock_open()
        with mock.patch("satpy.readers.ahi_l1b_gridded_bin.open", m, create=True):
            self.fh._download_luts("/test_file")
            mock_dl.assert_called()
            mock_shutil.assert_called()
